from io import StringIO
from mage_ai.io.base import ExportWritePolicy, QUERY_ROW_LIMIT
from mage_ai.io.config import BaseConfigLoader, ConfigKey
from mage_ai.io.export_utils import (
    clean_df_for_export,
    infer_dtypes,
)
from mage_ai.io.sql import BaseSQL
from mage_ai.shared.utils import (
    clean_name,
    convert_pandas_dtype_to_python_type,
    convert_python_type_to_trino_type,
)
from pandas import DataFrame, Series
from time import sleep
from trino.auth import BasicAuthentication
from trino.dbapi import Connection, Cursor as CursorParent
from trino.exceptions import TrinoUserError
from trino.transaction import IsolationLevel
from typing import IO, List, Mapping, Union


class Cursor(CursorParent):
    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        pass


class ConnectionWrapper(Connection):
    def cursor(self, legacy_primitive_types: bool = None):
        """Return a new :py:class:`Cursor` object using the connection."""
        if self.isolation_level != IsolationLevel.AUTOCOMMIT:
            if self.transaction is None:
                self.start_transaction()
        if self.transaction is not None:
            request = self.transaction.request
        else:
            request = self._create_request()
        return Cursor(
            self,
            request,
            legacy_primitive_types
            if legacy_primitive_types is not None
            else False
        )


class Trino(BaseSQL):
    QUERY_MAX_LENGTH = 100_000

    def __init__(
        self,
        catalog: str,
        host: str,
        user: str,
        password: str = None,
        port: int = 8080,
        schema: str = None,
        verbose: bool = True,
        **kwargs,
    ) -> None:
        super().__init__(
            verbose=verbose,
            catalog=catalog,
            host=host,
            user=user,
            password=password,
            port=port,
            schema=schema,
            **kwargs
        )

    @classmethod
    def with_config(cls, config: BaseConfigLoader) -> 'Trino':
        if config.get('trino'):
            settings = config['trino']

            return cls(
                catalog=settings.get('catalog'),
                host=settings.get('host'),
                http_headers=settings.get('http_headers'),
                http_scheme=settings.get('http_scheme'),
                password=settings.get('password'),
                port=settings.get('port'),
                schema=settings.get('schema'),
                session_properties=settings.get('session_properties'),
                source=settings.get('source'),
                user=settings.get('user'),
                verify=settings.get('verify'),
            )

        return cls(
            catalog=config[ConfigKey.TRINO_CATALOG],
            host=config[ConfigKey.TRINO_HOST],
            user=config[ConfigKey.TRINO_USER],
            password=config[ConfigKey.TRINO_PASSWORD],
            port=config[ConfigKey.TRINO_PORT],
            schema=config[ConfigKey.TRINO_SCHEMA],
        )

    def default_database(self) -> str:
        return self.settings.get('catalog')

    def default_schema(self) -> str:
        return self.settings.get('schema')

    def build_create_table_command(
        self,
        dtypes: Mapping[str, str],
        schema_name: str,
        table_name: str,
        unique_constraints: List[str] = [],
    ):
        query = []
        for cname in dtypes:
            query.append(f'"{clean_name(cname)}" {dtypes[cname]}')

        full_table_name = '.'.join(list(filter(lambda x: x, [
            schema_name,
            table_name,
        ])))

        return f'CREATE TABLE {full_table_name} (' + ','.join(query) + ')'

    def execute_queries(
        self,
        queries: List[str],
        **kwargs
    ) -> List:
        data = None
        tries = 0

        while data is None and tries < 3:
            if tries >= 1:
                sleep(1)

            try:
                data = super().execute_queries(queries, **kwargs)
            except TrinoUserError as err:
                print(err)

            tries += 1

        return data or []

    def load(
        self,
        query_string: str,
        **kwargs,
    ) -> DataFrame:
        data = None
        tries = 0

        while data is None and tries < 3:
            if tries >= 1:
                sleep(1)

            try:
                data = super().load(query_string, **kwargs)
            except TrinoUserError as err:
                print(err)

            tries += 1

        return data

    def open(self) -> None:
        with self.printer.print_msg('Opening connection to Trino database'):
            connect_kwargs = dict(
                catalog=self.settings.get('catalog'),
                host=self.settings.get('host'),
                http_headers=self.settings.get('http_headers'),
                http_scheme=self.settings.get('http_scheme'),
                port=self.settings.get('port'),
                schema=self.settings.get('schema'),
                session_properties=self.settings.get('session_properties'),
                source=self.settings.get('source'),
                user=self.settings.get('user'),
                verify=self.settings.get('verify'),
            )

            if self.settings.get('password'):
                connect_kwargs['auth'] = \
                    BasicAuthentication(
                        self.settings['user'], self.settings['password'])
                if 'http_scheme' not in connect_kwargs:
                    connect_kwargs['http_scheme'] = 'https'
            self._ctx = ConnectionWrapper(**connect_kwargs)

    def table_exists(self, schema_name: str, table_name: str) -> bool:
        with self.conn.cursor() as cur:
            catalog = self.default_database()

            cur.execute(f'SHOW SCHEMAS FROM {catalog} LIKE \'{schema_name}\'')
            if len(cur.fetchall()) == 0:
                return False

            cur.execute('\n'.join([
                f'SHOW TABLES FROM {catalog}.{schema_name} LIKE \'{table_name}\''
            ]))
            return len(cur.fetchall()) >= 1

    def upload_dataframe(
        self,
        cursor: Cursor,
        df: DataFrame,
        full_table_name: str,
        buffer: Union[IO, None] = None,
        **kwargs,
    ) -> None:
        values = []
        for _, row in df.iterrows():
            t = tuple(row)
            if len(t) == 1:
                values.append(f'({str(t[0])})')
            else:
                values.append(str(t))

        value_payload_size = 0
        subbatch_query = []
        for value in values:
            subbatch_query.append(value)
            value_payload_size += len(value) + 2
            if value_payload_size >= self.QUERY_MAX_LENGTH:
                values_string = ', '.join(subbatch_query)
                sql = f'INSERT INTO {full_table_name} VALUES {values_string}'
                cursor.execute(sql)
                subbatch_query = []
                value_payload_size = 0

        values_string = ', '.join(subbatch_query)
        sql = f'INSERT INTO {full_table_name} VALUES {values_string}'
        cursor.execute(sql)

    def get_type(self, column: Series, dtype: str) -> str:
        return convert_python_type_to_trino_type(
            convert_pandas_dtype_to_python_type(dtype)
        )

    def export(
        self,
        df: DataFrame,
        schema_name: str,
        table_name: str,
        if_exists: ExportWritePolicy = ExportWritePolicy.REPLACE,
        index: bool = False,
        verbose: bool = True,
        query_string: Union[str, None] = None,
        drop_table_on_replace: bool = False,
        cascade_on_drop: bool = False,
    ) -> None:
        """
        Exports dataframe to the connected database from a Pandas data frame. If table doesn't
        exist, the table is automatically created. If the schema doesn't exist, the schema is
        also created.

        Args:
            schema_name (str): Name of the schema of the table to export data to.
            table_name (str): Name of the table to insert rows from this data frame into.
            if_exists (ExportWritePolicy): Specifies export policy if table exists. Either
                - `'fail'`: throw an error.
                - `'replace'`: drops existing table and creates new table of same name.
                - `'append'`: appends data frame to existing table. In this case the schema must
                                match the original table.
            Defaults to `'replace'`.
            index (bool): If true, the data frame index is also exported alongside the table.
                            Defaults to False.
            **kwargs: Additional query parameters.
        """

        if type(df) is dict:
            df = DataFrame([df])
        elif type(df) is list:
            df = DataFrame(df)

        catalog = self.default_database()
        full_table_name = f'{catalog}.{schema_name}.{table_name}'

        if not query_string:
            if index:
                df = df.reset_index()

            dtypes = infer_dtypes(df)
            df = clean_df_for_export(df, self.clean, dtypes)

        def __process():
            buffer = StringIO()
            table_exists = self.table_exists(schema_name, table_name)

            with self.conn.cursor() as cur:
                if schema_name:
                    cur.execute(f'CREATE SCHEMA IF NOT EXISTS {catalog}.{schema_name}')

                should_create_table = not table_exists

                if table_exists:
                    if ExportWritePolicy.FAIL == if_exists:
                        raise ValueError(
                            f'Table \'{full_table_name}\' already exists in database.'
                        )
                    elif ExportWritePolicy.REPLACE == if_exists:
                        if drop_table_on_replace:
                            cmd = f'DROP TABLE {full_table_name}'
                            if cascade_on_drop:
                                cmd = f'{cmd} CASCADE'
                            cur.execute(cmd)
                            should_create_table = True
                        else:
                            cur.execute(f'DELETE FROM {full_table_name}')

                if query_string:
                    query = 'CREATE TABLE {} AS\n{}'.format(
                        full_table_name,
                        query_string,
                    )

                    if ExportWritePolicy.APPEND == if_exists and table_exists:
                        query = 'INSERT INTO {}\n{}'.format(
                            full_table_name,
                            query_string,
                        )
                    cur.execute(query)
                else:
                    if should_create_table:
                        db_dtypes = {
                            col: self.get_type(df[col], dtypes[col])
                            for col in dtypes
                        }
                        query = self.build_create_table_command(
                            db_dtypes,
                            schema_name,
                            table_name,
                        )
                        cur.execute(query)

                    self.upload_dataframe(cur, df, full_table_name, buffer)
            self.conn.commit()

        if verbose:
            with self.printer.print_msg(
                f'Exporting data to \'{full_table_name}\''
            ):
                __process()
        else:
            __process()

    def _enforce_limit(self, query: str, limit: int = QUERY_ROW_LIMIT) -> str:
        return f'SELECT * FROM ({query.strip(";")}) AS subquery LIMIT {limit}'



import json
import os
import shutil
import subprocess
from typing import Any, Dict, List

from mage_ai.data_preparation.models.block import Block
from mage_ai.data_preparation.models.block.dbt.utils import (
    build_command_line_arguments,
    compiled_query_string,
    create_temporary_profile,
    create_upstream_tables,
    fetch_model_data,
    load_profiles_async,
    parse_attributes,
    query_from_compiled_sql,
    run_dbt_tests,
    update_model_settings,
)
from mage_ai.data_preparation.models.constants import BlockLanguage, BlockType
from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.shared.hash import merge_dict


class DBTBlock(Block):
    @property
    def file_path(self) -> str:
        if BlockLanguage.SQL == self.language:
            repo_path = self.pipeline.repo_path if self.pipeline is not None else get_repo_path()
            file_path = self.configuration.get('file_path')

            return os.path.join(
                repo_path or os.getcwd(),
                'dbt',
                file_path,
            )

        return super().file_path

    async def metadata_async(self) -> Dict:
        project = None
        projects = {}
        block_metadata = {}

        if self.configuration.get('file_path'):
            attributes_dict = parse_attributes(self)
            profiles_full_path = attributes_dict['profiles_full_path']
            project_name = attributes_dict['project_name']
            profile_name = attributes_dict['profile_name']
            project = project_name

            targets = []
            profiles = await load_profiles_async(profile_name, profiles_full_path)
            outputs = profiles.get('outputs')
            if outputs:
                targets += sorted(list(outputs.keys()))

            projects[project_name] = dict(
                target=profiles.get('target'),
                targets=targets,
            )

            block_metadata['snapshot'] = attributes_dict['snapshot']
        else:
            dbt_dir = os.path.join(get_repo_path(), 'dbt')
            project_names = [
                name for name in os.listdir(dbt_dir) if os.path.isdir(os.path.join(dbt_dir, name))
            ]
            for project_name in project_names:
                profiles_full_path = os.path.join(dbt_dir, project_name, 'profiles.yml')
                targets = []
                profiles = await load_profiles_async(project_name, profiles_full_path)
                outputs = profiles.get('outputs')
                if outputs:
                    targets += sorted(list(outputs.keys()))

                projects[project_name] = dict(
                    target=profiles.get('target'),
                    targets=targets,
                )

        return dict(dbt=dict(
            block=block_metadata,
            project=project,
            projects=projects,
        ))

    def run_tests(
        self,
        build_block_output_stdout=None,
        global_vars=None,
        **kwargs,
    ):
        if self.configuration.get('file_path') is not None:
            attributes_dict = parse_attributes(self)
            snapshot = attributes_dict['snapshot']
            if snapshot:
                return

        run_dbt_tests(
            block=self,
            build_block_output_stdout=build_block_output_stdout,
            global_vars=global_vars,
        )

    def tags(self) -> List[str]:
        arr = super().tags()

        if self.configuration.get('file_path') is not None:
            attributes_dict = parse_attributes(self)
            if attributes_dict['snapshot']:
                from mage_ai.data_preparation.models.block.constants import (
                    TAG_DBT_SNAPSHOT,
                )
                arr.append(TAG_DBT_SNAPSHOT)

        return arr

    def update_upstream_blocks(self, upstream_blocks: List[Any]) -> None:
        upstream_blocks_previous = self.upstream_blocks
        super().update_upstream_blocks(upstream_blocks)
        if BlockLanguage.SQL == self.language:
            update_model_settings(self, upstream_blocks, upstream_blocks_previous)

    def _execute_block(
        self,
        outputs_from_input_vars,
        execution_partition: str = None,
        global_vars: Dict = None,
        test_execution: bool = False,
        runtime_arguments: Dict = None,
        run_settings: Dict = None,
        **kwargs,
    ) -> List:
        variables = merge_dict(global_vars, runtime_arguments or {})

        if run_settings:
            test_execution = not (
                run_settings.get('build_model', False) or
                run_settings.get('run_model', False) or
                run_settings.get('test_model', False)
            )

        dbt_command, args, command_line_dict = build_command_line_arguments(
            self,
            variables,
            run_settings=run_settings,
            test_execution=test_execution,
        )

        project_full_path = command_line_dict['project_full_path']
        profiles_dir = command_line_dict['profiles_dir']
        dbt_profile_target = command_line_dict['profile_target']

        # Create a temporary profiles file with variables and secrets interpolated
        _, temp_profile_full_path = create_temporary_profile(
            project_full_path,
            profiles_dir,
        )

        try:
            is_sql = BlockLanguage.SQL == self.language
            if is_sql:
                create_upstream_tables(
                    self,
                    execution_partition=execution_partition,
                    profile_target=dbt_profile_target,
                    # TODO (tommy dang): this is creating unnecessary tables in notebook
                    # cache_upstream_dbt_models=test_execution,
                )

            stdout = None if test_execution else subprocess.PIPE

            cmds = [
                'dbt',
                dbt_command,
            ] + args

            outputs = []

            args_pairs = []
            args_pair = []
            for a in args:
                args_pair.append(f"'{a}'" if '"' in a else a)
                if len(args_pair) == 2:
                    args_pairs.append(args_pair)
                    args_pair = []
            if len(args_pair) >= 1:
                args_pairs.append(args_pair)

            args_pairs = [f"  {' '.join(p)}" for p in args_pairs]
            args_string = ' \\\n'.join(args_pairs)

            print(f'dbt {dbt_command} \\\n{args_string}\n')

            snapshot = dbt_command == 'snapshot'

            if is_sql and test_execution:
                subprocess.run(
                    cmds,
                    preexec_fn=os.setsid,  # os.setsid doesn't work on Windows
                    stdout=stdout,
                )

                if not snapshot:
                    df = query_from_compiled_sql(
                        self,
                        dbt_profile_target,
                        limit=self.configuration.get('limit'),
                    )
                    self.store_variables(
                        dict(df=df),
                        execution_partition=execution_partition,
                        override_outputs=True,
                    )
                    outputs = [df]
            elif not test_execution:
                proc = subprocess.Popen(
                    cmds,
                    bufsize=1,
                    preexec_fn=os.setsid,  # os.setsid doesn't work on Windows
                    stdout=stdout,
                    universal_newlines=True,
                )
                for line in proc.stdout:
                    print(line, end='')
                proc.communicate()
                if proc.returncode != 0 and proc.returncode is not None:
                    raise subprocess.CalledProcessError(proc.returncode, proc.args)

            if not test_execution:
                target_path = None
                if self.configuration.get('file_path') is not None:
                    attributes_dict = parse_attributes(self)
                    target_path = attributes_dict['target_path']

                if snapshot:
                    query_string = compiled_query_string(self)
                    if query_string:

                        print('Compiled snapshot query string:')
                        for line in query_string.split('\n'):
                            print(f'|    {line.strip()}')
                elif target_path is not None:
                    run_results_file_path = os.path.join(
                        project_full_path,
                        target_path,
                        'run_results.json',
                    )
                    with open(run_results_file_path, 'r') as f:
                        try:
                            run_results = json.load(f)

                            print(f'\n{json.dumps(run_results, indent=2)}\n')

                            for result in run_results['results']:
                                if 'error' == result['status']:
                                    raise Exception(result['message'])
                        except json.decoder.JSONDecodeError:
                            print(f'WARNING: no run results found at {run_results_file_path}.')

                if is_sql and dbt_command in ['build', 'run']:
                    limit = 1000
                    if self.downstream_blocks and \
                            len(self.downstream_blocks) >= 1 and \
                            not all([BlockType.DBT == block.type
                                     for block in self.downstream_blocks]):
                        limit = None

                    try:
                        df = fetch_model_data(
                            self,
                            dbt_profile_target,
                            limit=limit,
                        )

                        self.store_variables(
                            dict(output_0=df),
                            execution_partition=execution_partition,
                            override_outputs=True,
                        )
                        outputs = [df]
                    except Exception as err:
                        print(f'Error: {err}')
        finally:
            # Always delete the temporary profiles dir
            try:
                shutil.rmtree(profiles_dir)
            except Exception as err:
                print(f'Error removing temporary profile at {temp_profile_full_path}: {err}')

        return outputs



from dataclasses import dataclass

from mage_ai.shared.config import BaseConfig


@dataclass
class CloudRunConfig(BaseConfig):
    project_id: str
    path_to_credentials_json_file: str = None
    region: str = 'us-west2'



import os

DEBUG = os.getenv('DEBUG', False)
HIDE_ENV_VAR_VALUES = int(os.getenv('HIDE_ENV_VAR_VALUES', 1) or 1) == 1
QUERY_API_KEY = 'api_key'

"""
import secrets
secrets.token_urlsafe()

Make sure this value is the same in mage_ai/frontend/api/constants.ts
"""
OAUTH2_APPLICATION_CLIENT_ID = 'zkWlN0PkIKSN0C11CfUHUj84OT5XOJ6tDZ6bDRO2'

# valid values: 0, 1, 2
try:
    DISABLE_NOTEBOOK_EDIT_ACCESS = int(os.getenv('DISABLE_NOTEBOOK_EDIT_ACCESS', 0))
except ValueError:
    DISABLE_NOTEBOOK_EDIT_ACCESS = 1 if os.getenv('DISABLE_NOTEBOOK_EDIT_ACCESS') else 0


def is_disable_pipeline_edit_access():
    return DISABLE_NOTEBOOK_EDIT_ACCESS >= 1


REQUIRE_USER_AUTHENTICATION = \
    os.getenv('REQUIRE_USER_AUTHENTICATION', 'False').lower() in ('true', '1', 't')
AUTHENTICATION_MODE = os.getenv('AUTHENTICATION_MODE', 'LOCAL')
LDAP_SERVER = os.getenv('LDAP_SERVER', 'ldaps://127.0.0.1:1636')
LDAP_BIND_DN = os.getenv('LDAP_BIND_DN', 'cd=admin,dc=example,dc=org')
LDAP_BIND_PASSWORD = os.getenv('LDAP_BIND_PASSWORD', 'admin_password')
LDAP_BASE_DN = os.getenv('LDAP_BASE_DN', 'dc=example,dc=org')
LDAP_AUTHENTICATION_FILTER = os.getenv('LDAP_AUTHENTICATION_FILTER',
                                       '(&(|(objectClass=Pers)(objectClass=gro))(cn={username}))')
LDAP_AUTHORIZATION_FILTER = os.getenv('LDAP_AUTHORIZATION_FILTER',
                                      '(&(objectClass=groupOfNames)(cn=group)(member={user_dn}))')
LDAP_ADMIN_USERNAME = os.getenv('LDAP_ADMIN_USERNAME', 'admin')

SERVER_VERBOSITY = os.getenv('SERVER_VERBOSITY', 'info') or 'info'

SHELL_COMMAND = os.getenv('SHELL_COMMAND', None)
USE_UNIQUE_TERMINAL = os.getenv('USE_UNIQUE_TERMINAL', None)

# sentry environment variables
SENTRY_DSN = os.getenv('SENTRY_DSN', None)
SENTRY_TRACES_SAMPLE_RATE = os.getenv('SENTRY_TRACES_SAMPLE_RATE', 1.0)

DEFAULT_LOCALHOST_URL = 'http://localhost:6789'
MAGE_PUBLIC_HOST = os.getenv('MAGE_PUBLIC_HOST') or DEFAULT_LOCALHOST_URL



from typing import Dict, List, Union

from pandas import DataFrame
from pymongo import MongoClient

from mage_ai.io.base import BaseIO
from mage_ai.io.config import BaseConfigLoader, ConfigKey


class MongoDB(BaseIO):
    def __init__(
        self,
        database: str,
        host: str,
        password: str,
        user: str,
        collection: str = None,
        port: int = 3306,
        verbose: bool = True,
        **kwargs,
    ) -> None:
        super().__init__(verbose=verbose)
        self.client = MongoClient(f'mongodb://{user}:{password}@{host}:{port}/')
        self.database = self.client[database]
        self.collection = collection

    @classmethod
    def with_config(cls, config: BaseConfigLoader) -> 'MongoDB':
        return cls(
            database=config[ConfigKey.MONGODB_DATABASE],
            host=config[ConfigKey.MONGODB_HOST],
            password=config[ConfigKey.MONGODB_PASSWORD],
            user=config[ConfigKey.MONGODB_USER],
            collection=config[ConfigKey.MONGODB_COLLECTION],
            port=config[ConfigKey.MONGODB_PORT],
        )

    def load(
        self,
        collection: str = None,
        query: Dict = dict(),
        **kwargs,
    ) -> DataFrame:
        """
        Loads the data frame from the MongoDB collection.

        Args:
            collection (str): MongoDB collection name.
            query (Dict): Filter the result by using a query object. Examples:
                { "address": "Park Lane 38" }, { "address": { "$gt": "S" } }

        Returns:
            DataFrame: Data frame object loaded from the MongoDB collection.
        """
        collection = collection or self.collection
        if collection is None:
            raise Exception(
                'Please provide the collection name either in the method args or in'
                ' io_config.yaml.')
        return DataFrame(list(self.database[collection].find(query)))

    def export(
        self,
        data: Union[DataFrame, List[Dict]],
        collection: str = None,
        **kwargs,
    ) -> None:
        """
        Exports the input dataframe to the MongoDB collection

        Args:
            data (Union[DataFrame, List[Dict]): Data frame or List of Dictionary to export.
            collection (str): MongoDB collection name.
        """
        if not data:
            return
        collection = collection or self.collection
        if collection is None:
            raise Exception(
                'Please provide the collection name either in the method args or in'
                ' io_config.yaml.')

        if type(data) is list and type(data[0]) is dict:
            records = data
        elif type(data) is DataFrame:
            records = data.to_dict('records')
        else:
            raise Exception('Please provide a pandas DataFrame or a list of dictionary as the'
                            ' input.')
        return self.database[collection].insert_many(records)



from json.decoder import JSONDecodeError
from mage_ai.api.errors import ApiError
from mage_ai.api.utils import authenticate_client_and_token, parse_cookie_header
from mage_ai.authentication.oauth2 import decode_token
from mage_ai.orchestration.db.models.oauth import Oauth2Application
from mage_ai.settings import OAUTH2_APPLICATION_CLIENT_ID, REQUIRE_USER_AUTHENTICATION
from mage_ai.server.api.constants import (
    COOKIE_OAUTH_TOKEN,
    ENDPOINTS_BYPASS_OAUTH_CHECK,
    HEADER_API_KEY,
    HEADER_OAUTH_TOKEN,
    URL_PARAMETER_API_KEY,
)
from mage_ai.shared.array import find
from tornado.web import RequestHandler
import json


class OAuthMiddleware(RequestHandler):
    def prepare(self):
        self.request.__setattr__('current_user', None)
        self.request.__setattr__('error', None)
        self.request.__setattr__('oauth_client', None)
        self.request.__setattr__('oauth_token', None)

        if not REQUIRE_USER_AUTHENTICATION:
            return

        paths = [path for path in self.request.uri.split('/') if path]
        if any(p in ENDPOINTS_BYPASS_OAUTH_CHECK for p in paths):
            return

        api_key = self.request.headers.get(HEADER_API_KEY, None)
        if not api_key:
            if self.request.method == 'POST':
                try:
                    api_key = json.loads(
                        self.request.body).get(URL_PARAMETER_API_KEY)
                except (JSONDecodeError, UnicodeDecodeError):
                    # UnicodeDecodeError happens when there is invalid characters in the
                    # file upload when creating a new feature set
                    api_key = self.request.query_arguments.get(
                        URL_PARAMETER_API_KEY)
            else:
                api_key = self.request.query_arguments.get(
                    URL_PARAMETER_API_KEY)

        if api_key:
            if isinstance(api_key, list):
                api_key = api_key[0]
            if not isinstance(api_key, str):
                api_key = api_key.decode()

        token_from_header = self.request.headers.get(HEADER_OAUTH_TOKEN, None)
        if not token_from_header:
            token_from_header = self.request.headers.get('Authorization')
            if token_from_header:
                tokens = token_from_header.split(',')
                token_from_header = find(lambda x: 'bearer' in x.lower(), tokens)
                if token_from_header:
                    token_from_header = (
                        token_from_header.
                        replace('Bearer ', '').
                        replace('bearer ', '')
                    )
                else:
                    token_from_header = None
            else:
                token_from_header = self.request.query_arguments.get('HTTP_AUTHORIZATION', None)

        cookies = parse_cookie_header(self.request.headers.get('Cookie', ''))

        if api_key:
            oauth_client = Oauth2Application.query_client(api_key)

            self.request.__setattr__('oauth_client', oauth_client)
            if not oauth_client:
                self.request.__setattr__('error', ApiError.INVALID_API_KEY)
            elif oauth_client.client_id != OAUTH2_APPLICATION_CLIENT_ID:
                self.request.__setattr__('error', ApiError.INVALID_API_KEY)
            else:
                should_check = False
                oauth_token = None
                if token_from_header:
                    oauth_token, valid = authenticate_client_and_token(
                        oauth_client.id,
                        token_from_header,
                    )
                    should_check = True
                elif COOKIE_OAUTH_TOKEN in cookies:
                    token_data = decode_token(cookies[COOKIE_OAUTH_TOKEN])
                    if 'token' in token_data:
                        oauth_token, valid = authenticate_client_and_token(
                            oauth_client.id,
                            decode_token(cookies[COOKIE_OAUTH_TOKEN])['token'],
                        )
                    should_check = True

                if should_check:
                    if oauth_token:
                        if valid:
                            self.request.__setattr__('oauth_token', oauth_token)
                            self.request.__setattr__(
                                'current_user', oauth_token.user)
                        else:
                            self.request.__setattr__(
                                'error', ApiError.EXPIRED_OAUTH_TOKEN)
                    else:
                        self.request.__setattr__(
                            'error', ApiError.INVALID_OAUTH_TOKEN)
        else:
            self.request.__setattr__('error', ApiError.INVALID_API_KEY)



from mage_ai.data_cleaner.transformer_actions.udf.base import BaseUDF


class Divide(BaseUDF):
    def execute(self):
        col1 = self.arguments[0]
        if len(self.arguments) > 1:
            col2 = self.arguments[1]
            return self.df[col1].astype(float) / self.df[col2].astype(float)
        elif self.options.get('value') is not None:
            return self.df[col1] / float(self.options['value'])
        raise Exception('Require second column or a value to divide.')



from mage_ai.api.oauth_scope import OauthScope
from mage_ai.api.operations import constants
from mage_ai.api.policies.BasePolicy import BasePolicy
from mage_ai.api.presenters.StatusPresenter import StatusPresenter


class StatusPolicy(BasePolicy):
    pass


StatusPolicy.allow_actions([
    constants.LIST,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
    OauthScope.CLIENT_PUBLIC,
])

StatusPolicy.allow_read(StatusPresenter.default_attributes, scopes=[
    OauthScope.CLIENT_PRIVATE,
    OauthScope.CLIENT_PUBLIC,
], on_action=[
    constants.LIST,
])



import os
from typing import Dict

from mage_ai.api.resources.GenericResource import GenericResource
from mage_ai.data_preparation.preferences import get_preferences
from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.data_preparation.shared.secrets import create_secret
from mage_ai.data_preparation.sync import (
    GIT_ACCESS_TOKEN_SECRET_NAME,
    GIT_SSH_PRIVATE_KEY_SECRET_NAME,
    GIT_SSH_PUBLIC_KEY_SECRET_NAME,
    GitConfig,
    UserGitConfig,
)
from mage_ai.data_preparation.sync.git_sync import GitSync
from mage_ai.orchestration.db import safe_db_query
from mage_ai.orchestration.db.models.oauth import User
from mage_ai.orchestration.db.models.secrets import Secret


def get_ssh_public_key_secret_name(user: User = None) -> str:
    return f'{GIT_SSH_PUBLIC_KEY_SECRET_NAME}_{user.id}' if user \
        else GIT_SSH_PUBLIC_KEY_SECRET_NAME


def get_ssh_private_key_secret_name(user: User = None) -> str:
    return f'{GIT_SSH_PRIVATE_KEY_SECRET_NAME}_{user.id}' if user \
        else GIT_SSH_PRIVATE_KEY_SECRET_NAME


def get_access_token_secret_name(user: User = None) -> str:
    return f'{GIT_ACCESS_TOKEN_SECRET_NAME}_{user.id}' if user \
        else GIT_ACCESS_TOKEN_SECRET_NAME


class SyncResource(GenericResource):
    @classmethod
    def collection(self, query, meta, user, **kwargs):
        sync_config = self.get_project_sync_config(user)
        return self.build_result_set(
            [sync_config],
            user,
            **kwargs,
        )

    @classmethod
    @safe_db_query
    def create(self, payload, user, **kwargs):
        user_settings = payload.pop('user_git_settings', dict())

        payload = self.update_user_settings(payload)
        preferences = get_preferences()
        updated_config = dict(preferences.sync_config, **payload)
        # default repo_path to os.getcwd()
        if not updated_config.get('repo_path', None):
            updated_config['repo_path'] = os.getcwd()

        # Update user git settings if they are included
        if user:
            # Validate payloads
            user_payload = self.update_user_settings(user_settings, user=user)
            UserGitConfig.load(config=user_payload)

            repo_path = get_repo_path()
            user_preferences = user.preferences or {}
            user_git_settings = user.git_settings or {}
            user_preferences[repo_path] = {
                **user_preferences.get(repo_path, {}),
                'git_settings': {
                    **user_git_settings,
                    **user_payload,
                }
            }
            user.refresh()
            user.update(preferences=user_preferences)

        sync_config = GitConfig.load(config=updated_config)

        preferences.update_preferences(dict(sync_config=updated_config))

        GitSync(sync_config)

        return self(get_preferences().sync_config, user, **kwargs)

    @classmethod
    def member(self, pk, user, **kwargs):
        sync_config = self.get_project_sync_config(user)
        return self(sync_config, user, **kwargs)

    def update(self, payload, **kwargs):
        self.model.pop('user_git_settings')
        config = GitConfig.load(config=self.model)
        sync = GitSync(config)
        action_type = payload.get('action_type')
        if action_type == 'sync_data':
            sync.sync_data()
        elif action_type == 'reset':
            sync.reset()

        return self

    @classmethod
    def get_project_sync_config(self, user):
        sync_config = get_preferences().sync_config
        # Make it backwards compatible with storing all of the git settings in the user
        # preferences field.
        if user and user.git_settings:
            sync_config['user_git_settings'] = user.git_settings
        else:
            sync_config['user_git_settings'] = UserGitConfig.from_dict(sync_config).to_dict()
        return sync_config

    @classmethod
    def update_user_settings(self, payload, user=None) -> Dict:
        user_payload = payload.copy()
        ssh_public_key = user_payload.pop('ssh_public_key', None)
        ssh_private_key = user_payload.pop('ssh_private_key', None)

        if ssh_public_key:
            secret_name = get_ssh_public_key_secret_name(user=user)
            secret = Secret.query.filter(
                Secret.name == secret_name).one_or_none()
            if secret:
                secret.delete()
            create_secret(secret_name, ssh_public_key)
            user_payload['ssh_public_key_secret_name'] = secret_name
        if ssh_private_key:
            secret_name = get_ssh_private_key_secret_name(user=user)
            secret = Secret.query.filter(
                Secret.name == secret_name).one_or_none()
            if secret:
                secret.delete()
            create_secret(secret_name, ssh_private_key)
            user_payload['ssh_private_key_secret_name'] = secret_name

        access_token = user_payload.pop('access_token', None)
        if access_token:
            secret_name = get_access_token_secret_name(user=user)
            secret = Secret.query.filter(
                Secret.name == secret_name).one_or_none()
            if secret:
                secret.delete()
            create_secret(secret_name, access_token)
            user_payload['access_token_secret_name'] = secret_name
        return user_payload



from mage_ai.api.presenters.BasePresenter import BasePresenter


class WorkspacePresenter(BasePresenter):
    default_attributes = [
        'name',
        'instance',
        'cluster_type',
        'access',
        'repo_path',
    ]



ENV_VAR_CONTAINER_GROUP_NAME = 'AZURE_CONTAINER_GROUP_NAME'
ENV_VAR_RESOURCE_GROUP_NAME = 'AZURE_RESOURCE_GROUP_NAME'
ENV_VAR_STORAGE_ACCOUNT_KEY = 'AZURE_STORAGE_ACCOUNT_KEY'
ENV_VAR_SUBSCRIPTION_ID = 'AZURE_SUBSCRIPTION_ID'



from mage_ai.api.oauth_scope import OauthScope
from mage_ai.api.operations import constants
from mage_ai.api.policies.BasePolicy import BasePolicy
from mage_ai.api.presenters.MonitorStatPresenter import MonitorStatPresenter


class MonitorStatPolicy(BasePolicy):
    pass


MonitorStatPolicy.allow_actions([
    constants.DETAIL,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.has_at_least_viewer_role())

MonitorStatPolicy.allow_read([] + MonitorStatPresenter.default_attributes, scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.DETAIL,
], condition=lambda policy: policy.has_at_least_viewer_role())

MonitorStatPolicy.allow_query([
    'end_time',
    'pipeline_schedule_id',
    'pipeline_uuid',
    'start_time',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.has_at_least_viewer_role())



from mage_ai.data_cleaner.transformer_actions.custom_action import execute_custom_action
from mage_ai.tests.base_test import TestCase
from pandas.testing import assert_frame_equal
import pandas as pd
import numpy as np


class CustomActionTests(TestCase):
    def test_decorated_custom_action(self):
        df = pd.DataFrame(
            [
                ['1', 1.000, '2021-10-01', 'Store 1', 23023],
                ['1', None, '2021-10-01', 'Store 2', np.nan],
                [np.nan, 1100, '', '', 90233],
                ['2', None, None, 'Store 1', 23920],
                ['2', 12.00, '2021-09-01', None, np.nan],
                ['2', 125.0, '2021-09-01', 'Store 3', 49833],
            ],
            columns=[
                'group_id',
                'price',
                'group_churned_at',
                'store',
                'zip_code',
            ],
        )
        expected_df = pd.DataFrame(
            [
                [1.0, 1.000, '2021-10-01', 'Store 1', 23023, 1],
                [1.0, None, '2021-10-01', 'Store 2', np.nan, 1],
                [np.nan, 1100, np.nan, np.nan, 90233, None],
                [2.0, None, None, 'Store 1', 23920, 2],
                [2.0, 12.00, '2021-09-01', None, np.nan, 2],
                [2.0, 125.0, '2021-09-01', 'Store 3', 49833, 2],
            ],
            columns=['group_id', 'price', 'group_churned_at', 'store', 'zip_code', 'new_col'],
        )
        action = dict(
            action_type='custom',
            axis='row',
            action_code='''from mage_ai.data_cleaner.column_types.column_type_detector \
            import infer_column_types
from mage_ai.data_cleaner.column_types.constants import ColumnType
from mage_ai.data_cleaner.transformer_actions.constants import CURRENCY_SYMBOLS
import pandas as pd
import numpy as np

def clean_series(series, column_type, dropna=True):
    series_cleaned = series.apply(lambda x: x.strip(" \\\'\\\"")
if type(x) is str else x)
    series_cleaned = series_cleaned.map(
        lambda x: x if (not isinstance(x, str) or (len(x) > 0 and
    not x.isspace())) else np.nan
    )
    if dropna:
        series_cleaned = series_cleaned.dropna()
    if series_cleaned.count() == 0:
        return series_cleaned
    first_item = series_cleaned.dropna().iloc[0]
    if column_type == ColumnType.NUMBER or column_type == ColumnType.NUMBER_WITH_DECIMALS:
        is_percent = False
        if type(first_item) is str:
            series_cleaned = series_cleaned.str.replace(",", "")
            if series_cleaned.str.count(CURRENCY_SYMBOLS).sum() != 0:
                series_cleaned = series_cleaned.str.replace(CURRENCY_SYMBOLS, "")
            elif series_cleaned.str.contains("%").sum() != 0:
                is_percent = True
                series_cleaned = series_cleaned.str.replace("%", "")
            series_cleaned = series_cleaned.str.replace(" ", "")
            if column_type == ColumnType.NUMBER:
                try:
                    series_cleaned = series_cleaned.astype(int)
                except ValueError:
                    series_cleaned = series_cleaned.astype(float)
            else:
                series_cleaned = series_cleaned.astype(float)
            if is_percent:
                series_cleaned /= 100
    elif column_type == ColumnType.DATETIME:
        series_cleaned = pd.to_datetime(series_cleaned, errors="coerce", infer_datetime_format=True)
    return series_cleaned

@transformer_action
def clean_df(df):
    ctypes = infer_column_types(df)
    for col in df.columns:
        df[col] = clean_series(df[col], ctypes, dropna=False)
    return df

@transformer_action
def print_df(df):
    print(df)
    return df

@transformer_action
def print_df_dtypes(df):
    print(df.dtypes)
    return df

@transformer_action
def add_column(df):
    columns = list(df.columns)
    df["new_col"] = df[columns[0]]
    return df
            ''',
            action_variables={},
        )
        new_df = execute_custom_action(df, action)
        new_df['group_id'] = new_df['group_id'].astype(float)
        new_df['new_col'] = new_df['new_col'].astype(float)
        assert_frame_equal(new_df, expected_df)

    def test_scripted_custom_action(self):
        df = pd.DataFrame(
            [
                ['1', 1.000, '2021-10-01', 'Store 1', 23023],
                ['1', None, '2021-10-01', 'Store 2', np.nan],
                [np.nan, 1100, '', '', 90233],
                ['2', None, None, 'Store 1', 23920],
                ['2', 12.00, '2021-09-01', None, np.nan],
                ['2', 125.0, '2021-09-01', 'Store 3', 49833],
            ],
            columns=[
                'group_id',
                'price',
                'group_churned_at',
                'store',
                'zip_code',
            ],
        )
        expected_df = pd.DataFrame(
            [
                [1.0, 1.000, '2021-10-01', 'Store 1', 23023, 1],
                [1.0, None, '2021-10-01', 'Store 2', np.nan, 1],
                [np.nan, 1100, np.nan, np.nan, 90233, None],
                [2.0, None, None, 'Store 1', 23920, 2],
                [2.0, 12.00, '2021-09-01', None, np.nan, 2],
                [2.0, 125.0, '2021-09-01', 'Store 3', 49833, 2],
            ],
            columns=['group_id', 'price', 'group_churned_at', 'store', 'zip_code', 'new_col'],
        )
        action = dict(
            action_type='custom',
            axis='row',
            action_code='''from mage_ai.data_cleaner.column_types.column_type_detector \
            import infer_column_types
from mage_ai.data_cleaner.column_types.constants import ColumnType
from mage_ai.data_cleaner.transformer_actions.constants import CURRENCY_SYMBOLS
import pandas as pd
import numpy as np

def clean_series(series, column_type, dropna=True):
    series_cleaned = series.apply(lambda x: x.strip(" \\\'\\\"")
if type(x) is str else x)
    series_cleaned = series_cleaned.map(
        lambda x: x if (not isinstance(x, str) or (len(x) > 0 and
    not x.isspace())) else np.nan
    )
    if dropna:
        series_cleaned = series_cleaned.dropna()
    if series_cleaned.count() == 0:
        return series_cleaned
    first_item = series_cleaned.dropna().iloc[0]
    if column_type == ColumnType.NUMBER or column_type == ColumnType.NUMBER_WITH_DECIMALS:
        is_percent = False
        if type(first_item) is str:
            series_cleaned = series_cleaned.str.replace(",", "")
            if series_cleaned.str.count(CURRENCY_SYMBOLS).sum() != 0:
                series_cleaned = series_cleaned.str.replace(CURRENCY_SYMBOLS, "")
            elif series_cleaned.str.contains("%").sum() != 0:
                is_percent = True
                series_cleaned = series_cleaned.str.replace("%", "")
            series_cleaned = series_cleaned.str.replace(" ", "")
            if column_type == ColumnType.NUMBER:
                try:
                    series_cleaned = series_cleaned.astype(int)
                except ValueError:
                    series_cleaned = series_cleaned.astype(float)
            else:
                series_cleaned = series_cleaned.astype(float)
            if is_percent:
                series_cleaned /= 100
    elif column_type == ColumnType.DATETIME:
        series_cleaned = pd.to_datetime(series_cleaned, errors="coerce", infer_datetime_format=True)
    return series_cleaned

ctypes = infer_column_types(df)
for col in df.columns:
    df[col] = clean_series(df[col], ctypes, dropna=False)
print(df)
print(df.dtypes)
columns = list(df.columns)
df["new_col"] = df[columns[0]]
            ''',
            action_variables={},
        )
        new_df = execute_custom_action(df, action)
        new_df['group_id'] = new_df['group_id'].astype(float)
        new_df['new_col'] = new_df['new_col'].astype(float)
        assert_frame_equal(new_df, expected_df)



from mage_ai.orchestration.triggers.loop_time_trigger import LoopTimeTrigger
from mage_ai.tests.base_test import TestCase
from unittest.mock import patch


class LoopTimeTriggerTests(TestCase):
    @patch('mage_ai.orchestration.triggers.time_trigger.check_sla')
    @patch('mage_ai.orchestration.triggers.time_trigger.schedule_all')
    def test_run(self, mock_schedule_all, mock_check_sla):
        trigger = LoopTimeTrigger()
        trigger.run()
        mock_schedule_all.assert_called_once()
        mock_check_sla.assert_called_once()



from mage_ai.api.presenters.BasePresenter import BasePresenter


class WidgetPresenter(BasePresenter):
    default_attributes = [
        'all_upstream_blocks_executed',
        'color',
        'configuration',
        'content',
        'downstream_blocks',
        'executor_config',
        'executor_type',
        'has_callback',
        'language',
        'name',
        'outputs',
        'status',
        'type',
        'upstream_blocks',
        'uuid',
    ]

    def present(self, **kwargs):
        if type(self.model) is dict:
            return self.model

        return self.model.to_dict(include_content=True)



CONNECTION_STATUS_ACTIVE = 'active'
CONNECTION_STATUS_DEPRECATED = 'deprecated'
CONNECTION_STATUS_INACTIVE = 'inactive'

JOB_STATUS_FAILED = 'failed'
JOB_STATUS_PENDING = 'pending'
JOB_STATUS_SUCCEEDED = 'succeeded'



from datetime import datetime, timedelta
from time import sleep
from typing import Dict, Optional

from mage_ai.api.resources.PipelineScheduleResource import PipelineScheduleResource
from mage_ai.data_integrations.utils.scheduler import initialize_state_and_runs
from mage_ai.data_preparation.models.pipeline import Pipeline
from mage_ai.data_preparation.models.triggers import ScheduleStatus, ScheduleType
from mage_ai.orchestration.db import db_connection
from mage_ai.orchestration.db.models.schedules import PipelineRun, PipelineSchedule
from mage_ai.orchestration.pipeline_scheduler import configure_pipeline_run_payload
from mage_ai.orchestration.triggers.constants import (
    DEFAULT_POLL_INTERVAL,
    TRIGGER_NAME_FOR_TRIGGER_CREATED_FROM_CODE,
)


def create_and_start_pipeline_run(
    pipeline: Pipeline,
    pipeline_schedule: PipelineSchedule,
    payload: Dict = None,
) -> PipelineRun:
    if payload is None:
        payload = {}
    configured_payload, is_integration = configure_pipeline_run_payload(
        pipeline_schedule,
        pipeline.type,
        payload,
    )

    pipeline_run = PipelineRun.create(**configured_payload)

    from mage_ai.orchestration.pipeline_scheduler import PipelineScheduler
    pipeline_scheduler = PipelineScheduler(pipeline_run)

    if is_integration:
        initialize_state_and_runs(
            pipeline_run,
            pipeline_scheduler.logger,
            pipeline_run.get_variables(),
        )
    pipeline_scheduler.start(should_schedule=False)

    return pipeline_run


def fetch_or_create_pipeline_schedule(pipeline: Pipeline) -> PipelineSchedule:
    pipeline_uuid = pipeline.uuid
    schedule_name = TRIGGER_NAME_FOR_TRIGGER_CREATED_FROM_CODE
    schedule_type = ScheduleType.API

    pipeline_schedule = PipelineSchedule.query.filter(
        PipelineSchedule.name == schedule_name,
        PipelineSchedule.pipeline_uuid == pipeline_uuid,
        PipelineSchedule.schedule_type == schedule_type,
    ).first()

    if not pipeline_schedule:
        resource = PipelineScheduleResource.create(
            dict(
                name=schedule_name,
                schedule_type=schedule_type,
                start_time=datetime.utcnow(),
                status=ScheduleStatus.ACTIVE,
            ),
            None,
            parent_model=pipeline,
        )
        pipeline_schedule = resource.model

    return pipeline_schedule


def check_pipeline_run_status(
    pipeline_run: PipelineRun,
    error_on_failure: bool = False,
    poll_interval: float = DEFAULT_POLL_INTERVAL,
    poll_timeout: Optional[float] = None,
    verbose: bool = True,
) -> PipelineRun:
    pipeline_uuid = pipeline_run.pipeline_uuid

    poll_start = datetime.now()
    while True:
        db_connection.session.refresh(pipeline_run)
        status = pipeline_run.status.value
        message = f'Pipeline run {pipeline_run.id} for pipeline {pipeline_uuid}: {status}.'

        if PipelineRun.PipelineRunStatus.FAILED.value == status:
            if error_on_failure:
                raise Exception(message)

        if verbose:
            print(message)

        if status in [
            PipelineRun.PipelineRunStatus.CANCELLED.value,
            PipelineRun.PipelineRunStatus.COMPLETED.value,
        ]:
            break

        if (
            poll_timeout
            and datetime.now()
            > poll_start + timedelta(seconds=poll_timeout)
        ):
            raise Exception(
                f'Pipeline run {pipeline_run.id} for pipeline {pipeline_uuid}: time out after '
                f'{datetime.now() - poll_start}. Last status was {status}.'
            )

        sleep(poll_interval)

    return pipeline_run


def trigger_pipeline(
    pipeline_uuid: str,
    variables: Dict = None,
    check_status: bool = False,
    error_on_failure: bool = False,
    poll_interval: float = DEFAULT_POLL_INTERVAL,
    poll_timeout: Optional[float] = None,
    verbose: bool = True,
) -> PipelineRun:
    if variables is None:
        variables = {}
    pipeline = Pipeline.get(pipeline_uuid)

    pipeline_schedule = fetch_or_create_pipeline_schedule(pipeline)

    pipeline_run = create_and_start_pipeline_run(
        pipeline,
        pipeline_schedule,
        dict(variables=variables),
    )

    if check_status:
        pipeline_run = check_pipeline_run_status(
            pipeline_run,
            error_on_failure=error_on_failure,
            poll_interval=poll_interval,
            poll_timeout=poll_timeout,
            verbose=verbose,
        )

    return pipeline_run



from dataclasses import dataclass
from mage_ai.shared.config import BaseConfig
from mage_ai.streaming.sinks.base import BaseSink
from typing import Dict, List
import time


@dataclass
class DummyConfig(BaseConfig):
    print_msg: bool = True


class DummySink(BaseSink):
    config_class = DummyConfig

    def write(self, data: Dict):
        if self.config.print_msg:
            self._print(f'Ingest data {data}, time={time.time()}')

    def batch_write(self, data: List[Dict]):
        if not data:
            return
        if self.config.print_msg:
            self._print(f'Batch ingest {len(data)} records, time={time.time()}. Sample: {data[0]}')



from mage_ai.data_cleaner.transformer_actions.constants import Operator
import re

ACTION_CODE_NAME = r'(?:([^\s()"\']+|[\'\"][^"\']+[\'\"]))'
ACTION_CODE_PATTERN = re.compile(
    rf'{ACTION_CODE_NAME} ([!=<>]+|(?:contains)|(?:not contains)) {ACTION_CODE_NAME}'
)
ORIGINAL_COLUMN_PREFIX = 'orig_'
QUOTES = '\"\''
TRANSFORMED_COLUMN_PREFIX = 'tf_'


def append_prefix(column_name, prefix):
    is_quoted = False
    if column_name[0] == '\"' or column_name[0] == '\'':
        is_quoted = True
        column_name = column_name.strip(QUOTES)
    column_name = f'{prefix}{column_name}'
    if is_quoted:
        column_name = f'`{column_name}`'
    return column_name


def __query_mutate_null_type(match, dtype):
    condition = ['']
    column_name, operator, _ = match.groups()
    column_name = append_prefix(column_name, ORIGINAL_COLUMN_PREFIX)
    if operator == '==':
        condition.append(f'({column_name}.isna()')
        if dtype == bool:
            condition.append(f' | {column_name} == \'\'')
        elif dtype == str:
            condition.append(f' | {column_name}.str.len() == 0')
        condition.append(')')
    else:
        condition.append(f'({column_name}.notna()')
        if dtype == bool:
            condition.append(f' & {column_name} != \'\'')
        elif dtype == str:
            condition.append(f' & {column_name}.str.len() >= 1')
        condition.append(')')
    return ''.join(condition)


def __query_mutate_contains_op(match):
    column_name, operator, value = match.groups()
    column_name = append_prefix(column_name, TRANSFORMED_COLUMN_PREFIX)
    value = value.strip(QUOTES)
    if operator == Operator.CONTAINS:
        condition = f'({column_name}.notna() & {column_name}.str.contains(\'{value}\'))'
    else:
        condition = f'~({column_name}.notna() & {column_name}.str.contains(\'{value}\'))'
    return condition


def __query_mutate_default_case(match, column_set):
    column_name, operator, value = match.groups()
    column_name = append_prefix(column_name, TRANSFORMED_COLUMN_PREFIX)
    if value.strip(QUOTES) in column_set:
        # if comparison is with another column, prefix value with column identifier
        value = append_prefix(value, TRANSFORMED_COLUMN_PREFIX)
    return f'{column_name} {operator} {value}'


def __get_column_type(df, cache, column_name):
    dtype = cache.get(column_name, None)
    if dtype is None:
        dropped_na = df[column_name].dropna()
        dropped_na = dropped_na[~dropped_na.isin([''])]
        dtype = type(dropped_na.iloc[0]) if len(dropped_na.index) >= 1 else object
        cache[column_name] = dtype
    return dtype


def query_with_action_code(df, action_code, kwargs):
    if action_code == '':
        return df

    transformed_types, original_types = {}, {}
    original_df, original_merged = kwargs.get('original_df', None), False
    reconstructed_code = []
    queried_df = df.copy().add_prefix(TRANSFORMED_COLUMN_PREFIX)
    column_set = set(df.columns)

    prev_end = 0
    for match in ACTION_CODE_PATTERN.finditer(action_code):
        column_name, operator, value = match.groups()
        column_name = column_name.strip(QUOTES)
        reconstructed_code.append(action_code[prev_end: match.start()])
        prev_end = match.end()
        if operator == Operator.CONTAINS or operator == Operator.NOT_CONTAINS:
            transformed_dtype = __get_column_type(df, transformed_types, column_name)
            if transformed_dtype != str:
                raise TypeError(
                    f'\'{operator}\' can only be used on string columns, {transformed_dtype}'
                )
            reconstructed_code.append(__query_mutate_contains_op(match))
        elif (operator == Operator.EQUALS or operator == Operator.NOT_EQUALS) and value == 'null':
            if original_df is None:
                raise Exception('Null value queries require original dataframe as keyword argument')
            elif not original_merged:
                queried_df = queried_df.join(original_df.add_prefix(ORIGINAL_COLUMN_PREFIX))
                original_merged = True
            original_dtype = __get_column_type(original_df, original_types, column_name)
            reconstructed_code.append(__query_mutate_null_type(match, original_dtype))
        else:
            reconstructed_code.append(__query_mutate_default_case(match, column_set))
    reconstructed_code.append(action_code[prev_end:])

    action_code = ''.join(reconstructed_code)
    queried_df = queried_df.query(action_code, engine='python').rename(
        lambda x: x[len(TRANSFORMED_COLUMN_PREFIX):], axis='columns'
    )
    return queried_df[df.columns]



from mage_ai.api.errors import ApiError
from mage_ai.api.resources.DatabaseResource import DatabaseResource
from mage_ai.data_preparation.preferences import get_preferences
from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.data_preparation.shared.secrets import (
    create_secret,
    get_valid_secrets,
)
from mage_ai.data_preparation.sync import (
    GitConfig,
    GIT_ACCESS_TOKEN_SECRET_NAME,
    GIT_SSH_PRIVATE_KEY_SECRET_NAME,
    GIT_SSH_PUBLIC_KEY_SECRET_NAME,
)
from mage_ai.orchestration.db import safe_db_query
from mage_ai.orchestration.db.models.secrets import Secret
from mage_ai.shared.hash import extract

ALLOWED_PAYLOAD_KEYS = [
    'name',
    'value',
]


class SecretResource(DatabaseResource):
    model_class = Secret

    @classmethod
    @safe_db_query
    def collection(self, query_arg, meta, user, **kwargs):
        secrets = get_valid_secrets()
        return list(filter(
            lambda s: self._filter_secrets(s, user),
            secrets
        ))

    @classmethod
    @safe_db_query
    def create(self, payload, user, **kwargs):
        return self(create_secret(**extract(payload, ALLOWED_PAYLOAD_KEYS)), user, **kwargs)

    @classmethod
    @safe_db_query
    def member(self, pk, user, **kwargs):
        repo_path = get_repo_path()
        model = Secret.query.filter(Secret.repo_name == repo_path, Secret.name == pk).first()

        if not model:
            raise ApiError(ApiError.RESOURCE_NOT_FOUND)

        return self(model, user, **kwargs)

    @classmethod
    def _filter_secrets(self, secret: Secret, user) -> bool:
        # Only include git secrets that were created by the current user.
        preferences = GitConfig(get_preferences(user=user).sync_config)
        whitelist_secrets = [
            preferences.ssh_private_key_secret_name,
            preferences.ssh_public_key_secret_name,
            preferences.access_token_secret_name,
        ]
        return not secret.name.startswith(GIT_SSH_PRIVATE_KEY_SECRET_NAME) \
            and not secret.name.startswith(GIT_SSH_PUBLIC_KEY_SECRET_NAME) \
            and not secret.name.startswith(GIT_ACCESS_TOKEN_SECRET_NAME) \
            or secret.name in whitelist_secrets



from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.io.config import ConfigFileLoader
from mage_ai.io.s3 import S3
from os import path

if 'sensor' not in globals():
    from mage_ai.data_preparation.decorators import sensor


@sensor
def check_condition(*args, **kwargs) -> bool:
    """
    Template code for checking if a file or folder exists in a S3 bucket

    You will also need to fill out the following AWS related fields
    in `io_config.yaml`:
        - AWS_ACCESS_KEY_ID
        - AWS_SECRET_ACCESS_KEY
        - AWS_REGION
    """

    config_path = path.join(get_repo_path(), 'io_config.yaml')
    config_profile = 'default'

    bucket_name = 'your_bucket_name'
    s3_path = 'path/to/folder/or/file'

    config_file_loader = ConfigFileLoader(config_path, config_profile)
    return S3.with_config(config_file_loader).exists(
        bucket_name, s3_path
    )



import os
import time

from azure.identity import DefaultAzureCredential
from azure.mgmt.containerinstance import ContainerInstanceManagementClient
from azure.mgmt.containerinstance.models import (
    Container,
    ContainerGroup,
    ContainerGroupRestartPolicy,
    ResourceRequests,
    ResourceRequirements,
)

from mage_ai.server.logger import Logger
from mage_ai.services.azure.constants import (
    ENV_VAR_CONTAINER_GROUP_NAME,
    ENV_VAR_RESOURCE_GROUP_NAME,
    ENV_VAR_STORAGE_ACCOUNT_KEY,
    ENV_VAR_SUBSCRIPTION_ID,
)
from mage_ai.services.azure.container_instance.config import ContainerInstanceConfig

logger = Logger().new_server_logger(__name__)


def run_job(command: str, job_id: str, container_instance_config: ContainerInstanceConfig):
    if type(container_instance_config) is dict:
        container_instance_config = ContainerInstanceConfig.load(config=container_instance_config)

    curr_container_group_name = os.getenv(ENV_VAR_CONTAINER_GROUP_NAME)
    resource_group_name = os.getenv(ENV_VAR_RESOURCE_GROUP_NAME)
    subscription_id = os.getenv(ENV_VAR_SUBSCRIPTION_ID)

    credential = DefaultAzureCredential()

    # Create the ACI management client
    aci_client = ContainerInstanceManagementClient(credential, subscription_id)

    # Get current container group
    curr_container_group = aci_client.container_groups.get(
        resource_group_name,
        curr_container_group_name,
    )
    curr_container = curr_container_group.containers[0]

    # Configure the container
    container_resource_requests = ResourceRequests(
        cpu=container_instance_config.cpu,
        memory_in_gb=container_instance_config.memory,
    )
    container_resource_requirements = ResourceRequirements(
        requests=container_resource_requests,
    )
    container = Container(
        command=command.split(' '),
        environment_variables=curr_container.environment_variables,
        image=curr_container.image,
        name=job_id,
        resources=container_resource_requirements,
        volume_mounts=curr_container.volume_mounts,
    )

    # Configure the container group
    volumes = []
    if curr_container_group.volumes:
        volume = curr_container_group.volumes[0]
        volume.azure_file.storage_account_key = os.getenv(ENV_VAR_STORAGE_ACCOUNT_KEY)
        volumes = [volume]
    group = ContainerGroup(
        containers=[container],
        location=curr_container_group.location,
        os_type=curr_container_group.os_type,
        restart_policy=ContainerGroupRestartPolicy.never,
        volumes=volumes,
    )

    # Create the container group
    result = aci_client.container_groups.begin_create_or_update(
        resource_group_name,
        job_id,
        group,
    )
    logger.info(f'Begin creating or updating the job {job_id}')

    # Wait for the container create operation to complete. The operation is
    # "done" when the container group provisioning state is one of:
    # Succeeded, Canceled, Failed
    while result.done() is False:
        time.sleep(5)

    # Get the provisioning state of the container group.
    container_group = aci_client.container_groups.get(resource_group_name,
                                                      job_id)
    if str(container_group.provisioning_state).lower() == 'succeeded':
        logger.info(f'Executing job {job_id} succeeded.')
    else:
        logger.info(f'Executing job {job_id} failed.'
                    f'Provisioning state is: {container_group.provisioning_state}')



from enum import Enum

ENV_DEV = 'dev'
ENV_PROD = 'prod'
ENV_STAGING = 'staging'
ENV_TEST = 'test'

SAMPLE_SIZE = 1000

S3_PREFIX = 's3://'


class InstanceType(str, Enum):
    SERVER_AND_SCHEDULER = 'server_and_scheduler'
    SCHEDULER = 'scheduler'
    WEB_SERVER = 'web_server'



import asyncio

from sqlalchemy.orm import aliased

from mage_ai.api.operations.constants import DELETE
from mage_ai.api.resources.BaseResource import BaseResource
from mage_ai.data_preparation.models.block.dbt.utils import (
    add_blocks_upstream_from_refs,
)
from mage_ai.data_preparation.models.constants import PipelineStatus
from mage_ai.data_preparation.models.pipeline import Pipeline
from mage_ai.data_preparation.models.triggers import ScheduleStatus
from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.orchestration.db import safe_db_query
from mage_ai.orchestration.db.models.schedules import PipelineRun, PipelineSchedule
from mage_ai.orchestration.pipeline_scheduler import (
    PipelineScheduler,
    retry_pipeline_run,
)
from mage_ai.server.active_kernel import switch_active_kernel
from mage_ai.server.kernels import PIPELINE_TO_KERNEL_NAME
from mage_ai.shared.hash import group_by, ignore_keys
from mage_ai.usage_statistics.logger import UsageStatisticLogger


class PipelineResource(BaseResource):
    @classmethod
    @safe_db_query
    async def collection(self, query, meta, user, **kwargs):
        include_schedules = query.get('include_schedules', [False])
        if include_schedules:
            include_schedules = include_schedules[0]

        pipeline_types = query.get('type[]', [])
        if pipeline_types:
            pipeline_types = pipeline_types[0]
        if pipeline_types:
            pipeline_types = pipeline_types.split(',')

        pipeline_statuses = query.get('status[]', [])
        if pipeline_statuses:
            pipeline_statuses = pipeline_statuses[0]
        if pipeline_statuses:
            pipeline_statuses = pipeline_statuses.split(',')

        pipeline_uuids = Pipeline.get_all_pipelines(get_repo_path())

        await UsageStatisticLogger().pipelines_impression(lambda: len(pipeline_uuids))

        async def get_pipeline(uuid):
            try:
                return await Pipeline.get_async(uuid)
            except Exception as err:
                err_message = f'Error loading pipeline {uuid}: {err}.'
                if err.__class__.__name__ == 'OSError' and 'Too many open files' in err.strerror:
                    raise Exception(err_message)
                else:
                    print(err_message)
                    return None

        pipelines = await asyncio.gather(
            *[get_pipeline(uuid) for uuid in pipeline_uuids]
        )
        pipelines = [p for p in pipelines if p is not None]

        @safe_db_query
        def query_pipeline_schedules(pipeline_uuids):
            a = aliased(PipelineSchedule, name='a')
            result = (
                PipelineSchedule.
                select(*[
                    a.created_at,
                    a.id,
                    a.name,
                    a.pipeline_uuid,
                    a.schedule_interval,
                    a.schedule_type,
                    a.start_time,
                    a.status,
                    a.updated_at,
                ]).
                filter(a.pipeline_uuid.in_(pipeline_uuids))
            ).all()
            return group_by(lambda x: x.pipeline_uuid, result)

        mapping = {}
        if include_schedules:
            mapping = query_pipeline_schedules(pipeline_uuids)
        if pipeline_types:
            pipelines = [p for p in pipelines if p.type in pipeline_types]

        filtered_pipelines = []
        for pipeline in pipelines:
            schedules = []
            if mapping.get(pipeline.uuid):
                schedules = mapping[pipeline.uuid]
            pipeline.schedules = schedules

            if pipeline_statuses and (
                (PipelineStatus.ACTIVE in pipeline_statuses and
                    any(s.status == ScheduleStatus.ACTIVE
                        for s in pipeline.schedules)) or
                (PipelineStatus.INACTIVE in pipeline_statuses and
                    len(pipeline.schedules) > 0 and
                    all(s.status == ScheduleStatus.INACTIVE
                        for s in pipeline.schedules)) or
                (PipelineStatus.NO_SCHEDULES in pipeline_statuses and
                    len(pipeline.schedules) == 0)
            ):
                filtered_pipelines.append(pipeline)

        if include_schedules and pipeline_statuses:
            pipelines = filtered_pipelines

        return self.build_result_set(
            pipelines,
            user,
            **kwargs,
        )

    @classmethod
    @safe_db_query
    def create(self, payload, user, **kwargs):
        clone_pipeline_uuid = payload.get('clone_pipeline_uuid')
        name = payload.get('name')
        pipeline_type = payload.get('type')

        if clone_pipeline_uuid is None:
            pipeline = Pipeline.create(
                name,
                pipeline_type=pipeline_type,
                repo_path=get_repo_path(),
            )
        else:
            source = Pipeline.get(clone_pipeline_uuid)
            pipeline = Pipeline.duplicate(source, name)

        return self(pipeline, user, **kwargs)

    @classmethod
    @safe_db_query
    async def get_model(self, pk):
        return await Pipeline.get_async(pk)

    @classmethod
    @safe_db_query
    async def member(self, pk, user, **kwargs):
        pipeline = await Pipeline.get_async(pk)

        if kwargs.get('api_operation_action', None) != DELETE:
            switch_active_kernel(PIPELINE_TO_KERNEL_NAME[pipeline.type])

        return self(pipeline, user, **kwargs)

    @safe_db_query
    def delete(self, **kwargs):
        return self.model.delete()

    @safe_db_query
    async def update(self, payload, **kwargs):
        if 'add_upstream_for_block_uuid' in payload:
            block_uuid = payload['add_upstream_for_block_uuid']
            block = self.model.get_block(block_uuid, widget=False)
            arr = add_blocks_upstream_from_refs(block)
            upstream_block_uuids = [b.uuid for b in arr]

            for b in block.upstream_blocks:
                upstream_block_uuids.append(b.uuid)

            self.model.add_block(
                block,
                upstream_block_uuids,
                priority=len(upstream_block_uuids),
                widget=False,
            )

            return self

        query = kwargs.get('query', {})
        update_content = query.get('update_content', [False])
        if update_content:
            update_content = update_content[0]

        await self.model.update(
            ignore_keys(payload, ['add_upstream_for_block_uuid']),
            update_content=update_content,
        )
        switch_active_kernel(PIPELINE_TO_KERNEL_NAME[self.model.type])

        @safe_db_query
        def update_schedule_status(status, pipeline_uuid):
            schedules = (
                PipelineSchedule.
                query.
                filter(PipelineSchedule.pipeline_uuid == pipeline_uuid)
            ).all()
            for schedule in schedules:
                schedule.update(status=status)

        @safe_db_query
        def cancel_pipeline_runs(status, pipeline_uuid):
            pipeline_runs = (
                PipelineRun.
                query.
                filter(PipelineRun.pipeline_uuid == pipeline_uuid).
                filter(PipelineRun.status.in_([
                    PipelineRun.PipelineRunStatus.INITIAL,
                    PipelineRun.PipelineRunStatus.RUNNING,
                ]))
            )
            for pipeline_run in pipeline_runs:
                PipelineScheduler(pipeline_run).stop()

        def retry_pipeline_runs(pipeline_runs):
            for run in pipeline_runs:
                retry_pipeline_run(run)

        status = payload.get('status')
        pipeline_uuid = self.model.uuid

        def _update_callback(resource):
            if status:
                if status in [
                    ScheduleStatus.ACTIVE.value,
                    ScheduleStatus.INACTIVE.value,
                ]:
                    update_schedule_status(status, pipeline_uuid)
                elif status == PipelineRun.PipelineRunStatus.CANCELLED.value:
                    cancel_pipeline_runs(status, pipeline_uuid)
                elif status == 'retry' and payload.get('pipeline_runs'):
                    retry_pipeline_runs(payload.get('pipeline_runs'))

        self.on_update_callback = _update_callback

        return self



from logging import Logger
from mage_ai.extensions.great_expectations.constants import (
    EXTENSION_UUID as EXTENSION_UUID_GREAT_EXPECTATIONS,
)
from mage_ai.shared.hash import index_by
from typing import Dict


def handle_run_tests(
    block,
    dynamic_block_uuid: str = None,
    execution_partition: str = None,
    global_vars: Dict = {},
    logger: Logger = None,
    logging_tags: Dict = {},
):
    if not block.pipeline:
        return

    extensions = block.pipeline.extensions
    if EXTENSION_UUID_GREAT_EXPECTATIONS not in extensions:
        return

    extension = extensions[EXTENSION_UUID_GREAT_EXPECTATIONS]
    blocks_by_uuid = extension.get('blocks_by_uuid', {})

    extension_blocks = []
    for extension_block in blocks_by_uuid.values():
        upstream_blocks_by_uuid = index_by(lambda x: x.uuid, extension_block.upstream_blocks)
        if block.uuid in upstream_blocks_by_uuid:
            extension_block.upstream_blocks = [block]
            extension_blocks.append(extension_block)

    for extension_block in extension_blocks:
        extension_block.execute_sync(
            dynamic_block_uuid=dynamic_block_uuid,
            execution_partition=execution_partition,
            global_vars=global_vars,
            logger=logger,
            logging_tags=logging_tags,
        )



import os
import shutil
import sys
import traceback
from enum import Enum
from typing import Dict

import ruamel.yaml
import yaml
from jinja2 import Template

from mage_ai.data_preparation.shared.constants import (
    MAGE_DATA_DIR_ENV_VAR,
    REPO_PATH_ENV_VAR,
)
from mage_ai.data_preparation.templates.utils import copy_template_directory
from mage_ai.shared.environments import is_test


if is_test():
    DEFAULT_MAGE_DATA_DIR = './'
else:
    DEFAULT_MAGE_DATA_DIR = '~/.mage_data'


class ProjectType(str, Enum):
    MAIN = 'main'
    SUB = 'sub'
    STANDALONE = 'standalone'


class RepoConfig:
    def __init__(self, repo_path: str = None, config_dict: Dict = None):
        from mage_ai.data_preparation.shared.utils import get_template_vars
        self.repo_path = repo_path or get_repo_path()
        self.repo_name = os.path.basename(self.repo_path)
        try:
            if not config_dict:
                if os.path.exists(self.metadata_path):
                    with open(self.metadata_path) as f:
                        config_file = Template(f.read()).render(
                            **get_template_vars()
                        )
                        repo_config = yaml.full_load(config_file) or {}
                else:
                    repo_config = dict()
            else:
                repo_config = config_dict

            # Priority:
            # 1. 'variables_dir' from config_dict
            # 1. os.getenv(MAGE_DATA_DIR_ENV_VAR)
            # 2. 'variables_dir' from project's metadata.yaml file
            # 3. DEFAULT_MAGE_DATA_DIR
            if config_dict is not None and config_dict.get('variables_dir'):
                self.variables_dir = config_dict.get('variables_dir')
            elif os.getenv(MAGE_DATA_DIR_ENV_VAR):
                self.variables_dir = os.getenv(MAGE_DATA_DIR_ENV_VAR)
            else:
                self.variables_dir = os.path.expanduser(
                    repo_config.get('variables_dir', DEFAULT_MAGE_DATA_DIR),
                )
            if self.variables_dir is not None and not self.variables_dir.startswith('s3'):
                if os.path.isabs(self.variables_dir) and self.variables_dir != self.repo_path and (
                    not config_dict or not config_dict.get('variables_dir')
                ):
                    # If the variables_dir is an absolute path, not same as repo_path, and
                    # from config file
                    self.variables_dir = os.path.join(self.variables_dir, self.repo_name)
                else:
                    self.variables_dir = os.path.abspath(
                        os.path.join(self.repo_path, self.variables_dir),
                    )
            os.makedirs(self.variables_dir, exist_ok=True)

            self.project_type = repo_config.get('project_type')
            self.cluster_type = repo_config.get('cluster_type')
            self.remote_variables_dir = repo_config.get('remote_variables_dir')

            # Executor configs
            self.azure_container_instance_config = \
                repo_config.get('azure_container_instance_config')
            self.ecs_config = repo_config.get('ecs_config')
            self.emr_config = repo_config.get('emr_config')
            self.gcp_cloud_run_config = repo_config.get('gcp_cloud_run_config')
            self.k8s_executor_config = repo_config.get('k8s_executor_config')
            self.spark_config = repo_config.get('spark_config')

            self.notification_config = repo_config.get('notification_config', dict())
            self.queue_config = repo_config.get('queue_config', dict())
            self.project_uuid = repo_config.get('project_uuid')
            self.help_improve_mage = repo_config.get('help_improve_mage')
            self.retry_config = repo_config.get('retry_config')

            self.s3_bucket = None
            self.s3_path_prefix = None
            if self.remote_variables_dir is not None and \
                    self.remote_variables_dir.startswith('s3://'):
                path_parts = self.remote_variables_dir.replace('s3://', '').split('/')
                self.s3_bucket = path_parts.pop(0)
                self.s3_path_prefix = '/'.join(path_parts)

            self.logging_config = repo_config.get('logging_config', dict())

            self.variables_retention_period = repo_config.get('variables_retention_period')
        except Exception:
            traceback.print_exc()
            pass

    @classmethod
    def from_dict(self, config_dict: Dict) -> 'RepoConfig':
        repo_path = config_dict.get('repo_path')
        repo_config = RepoConfig(repo_path=repo_path, config_dict=config_dict)
        return repo_config

    @property
    def metadata_path(self) -> str:
        return os.path.join(self.repo_path, 'metadata.yaml')

    def to_dict(self, remote: bool = False) -> Dict:
        return dict(
            project_type=self.project_type,
            azure_container_instance_config=self.azure_container_instance_config,
            ecs_config=self.ecs_config,
            emr_config=self.emr_config,
            gcp_cloud_run_config=self.gcp_cloud_run_config,
            notification_config=self.notification_config,
            queue_config=self.queue_config,
            repo_path=self.repo_path,
            variables_dir=self.remote_variables_dir if remote else self.variables_dir,
            variables_retention_period=self.variables_retention_period,
            remote_variables_dir=self.remote_variables_dir,
            project_uuid=self.project_uuid,
            help_improve_mage=self.help_improve_mage,
            spark_config=self.spark_config,
        )

    def save(self, **kwargs) -> None:
        yml = ruamel.yaml.YAML()
        yml.preserve_quotes = True
        yml.indent(mapping=2, sequence=2, offset=0)

        if os.path.exists(self.metadata_path):
            with open(self.metadata_path) as f:
                data = yml.load(f)
        else:
            data = {}

        for key, value in kwargs.items():
            data[key] = value
            if hasattr(self, key):
                setattr(self, key, value)

        with open(self.metadata_path, 'w') as f:
            yml.dump(data, f)


def init_repo(repo_path: str, project_type: str = ProjectType.STANDALONE) -> None:
    """
    Initialize a repository under the current path.
    """
    if os.path.exists(repo_path):
        raise FileExistsError(f'Repository {repo_path} already exists')

    if project_type == ProjectType.MAIN:
        copy_template_directory('main', repo_path)
    elif project_type == ProjectType.SUB:
        os.makedirs(
            os.getenv(MAGE_DATA_DIR_ENV_VAR) or DEFAULT_MAGE_DATA_DIR,
            exist_ok=True,
        )
        copy_template_directory('repo', repo_path)
        new_repo_config = get_repo_config(repo_path)
        current_metadata = get_repo_config().metadata_path
        new_metadata = new_repo_config.metadata_path
        if os.path.exists(current_metadata):
            shutil.copyfile(current_metadata, new_metadata)
        new_repo_config.save(
            project_type=ProjectType.SUB.value,
        )
    else:
        os.makedirs(
            os.getenv(MAGE_DATA_DIR_ENV_VAR) or DEFAULT_MAGE_DATA_DIR,
            exist_ok=True,
        )
        copy_template_directory('repo', repo_path)


def get_data_dir() -> str:
    return os.getenv(MAGE_DATA_DIR_ENV_VAR) or DEFAULT_MAGE_DATA_DIR


def get_repo_name() -> str:
    return os.path.basename(get_repo_path())


def get_repo_path() -> str:
    return os.getenv(REPO_PATH_ENV_VAR) or os.getcwd()


def get_repo_config(repo_path=None) -> RepoConfig:
    return RepoConfig(repo_path=repo_path)


def get_project_type(repo_path=None) -> ProjectType:
    return get_repo_config(repo_path=repo_path).project_type


def set_repo_path(repo_path: str) -> None:
    os.environ[REPO_PATH_ENV_VAR] = repo_path
    sys.path.append(os.path.dirname(repo_path))


def get_variables_dir(repo_path: str = None) -> str:
    return get_repo_config(repo_path=repo_path).variables_dir



from mage_ai.services.teams.config import TeamsConfig
import requests


def send_teams_message(
    config: TeamsConfig,
    message: str,
    title: str = 'Mage pipeline run status logs',
) -> None:
    requests.post(
        url=config.webhook_url,
        json={
            'summary': title,
            'sections': [{
                'activityTitle': title,
                'activitySubtitle': message,
            }],
        },
    )



import os

MIN_SECRET_ENV_VAR_LENGTH = 8
WHITELISTED_ENV_VARS = set([
    'MAGE_DATA_DIR',
    'MAGE_REPO_PATH',
    'HOME',
    'PWD',
    'PYTHONPATH',
])


def filter_out_env_var_values(value: str):
    env_var_values = dict(os.environ).values()
    whitelisted_env_var_values = set([os.getenv(k) for k in WHITELISTED_ENV_VARS if os.getenv(k)])
    env_var_values = [v for v in env_var_values
                      if v and len(v) >= MIN_SECRET_ENV_VAR_LENGTH
                      and v not in whitelisted_env_var_values]
    env_var_values.sort(key=len, reverse=True)
    value_clean = value
    for env_var_value in env_var_values:
        replace_value = '*' * len(env_var_value)
        value_clean = value_clean.replace(env_var_value, replace_value)
    return value_clean



from dataclasses import dataclass

from mage_ai.shared.config import BaseConfig


@dataclass
class ContainerInstanceConfig(BaseConfig):
    cpu: int = 1        # Number of CPU cores
    memory: int = 1     # Memory in GB



from enum import Enum
from dateutil.relativedelta import relativedelta


VARIABLE_NAME_BUCKETS = 'buckets'
VARIABLE_NAME_GROUP_BY = 'group_by'
VARIABLE_NAME_INDEX = 'index'
VARIABLE_NAME_LIMIT = 'limit'
VARIABLE_NAME_METRICS = 'metrics'
VARIABLE_NAME_TIME_INTERVAL = 'time_interval'
VARIABLE_NAME_X = 'x'
VARIABLE_NAME_Y = 'y'


class AggregationFunction(str, Enum):
    AVERAGE = 'average'
    COUNT = 'count'
    COUNT_DISTINCT = 'count_distinct'
    MAX = 'max'
    MEDIAN = 'median'
    MIN = 'min'
    MODE = 'mode'
    SUM = 'sum'


class TimeInterval(str, Enum):
    DAY = 'day'
    HOUR = 'hour'
    MINUTE = 'minute'
    MONTH = 'month'
    ORIGINAL = 'original'
    SECOND = 'second'
    WEEK = 'week'
    YEAR = 'year'


class ChartType(str, Enum):
    BAR_CHART = 'bar chart'
    HISTOGRAM = 'histogram'
    LINE_CHART = 'line chart'
    PIE_CHART = 'pie chart'
    TABLE = 'table'
    TIME_SERIES_BAR_CHART = 'time series bar chart'
    TIME_SERIES_LINE_CHART = 'time series line chart'


TIME_INTERVAL_TO_TIME_DELTA = {
    TimeInterval.DAY: relativedelta(days=1),
    TimeInterval.HOUR: relativedelta(hours=1),
    TimeInterval.MINUTE: relativedelta(minutes=1),
    TimeInterval.MONTH: relativedelta(months=1),
    TimeInterval.SECOND: relativedelta(seconds=1),
    TimeInterval.WEEK: relativedelta(weeks=1),
    TimeInterval.YEAR: relativedelta(years=1),
}


VARIABLE_NAMES_BY_CHART_TYPE = {
    ChartType.BAR_CHART: [
        VARIABLE_NAME_X,
        VARIABLE_NAME_Y,
    ],
    ChartType.HISTOGRAM: [
        VARIABLE_NAME_X,
    ],
    ChartType.LINE_CHART: [
        VARIABLE_NAME_X,
        VARIABLE_NAME_Y,
    ],
    ChartType.PIE_CHART: [
        VARIABLE_NAME_X,
    ],
    ChartType.TABLE: [
        VARIABLE_NAME_INDEX,
        VARIABLE_NAME_X,
        VARIABLE_NAME_Y,
    ],
}



import os
import traceback
from dataclasses import asdict, dataclass, is_dataclass
from typing import Dict

import yaml

from mage_ai.shared.hash import merge_dict
from mage_ai.shared.strings import camel_to_snake_case


@dataclass
class BaseConfig:
    @classmethod
    def load(self, config_path: str = None, config: Dict = None):
        config_class_name = self.__name__
        config_class_key = camel_to_snake_case(config_class_name)
        if config is None:
            self.config_path = config_path
            if self.config_path is None:
                raise Exception(
                    'Please provide a config_path or a config dictionary to initialize'
                    f' an {config_class_name} object',
                )
            if not os.path.exists(self.config_path):
                raise Exception(f'{config_class_name} {self.config_path} does not exist.')
            with open(self.config_path) as fp:
                config = yaml.full_load(fp) or {}

        if config_class_key in config:
            config = config[config_class_key]

        config = self.parse_config(config)
        extra_config = self.load_extra_config()
        config = merge_dict(config, extra_config)

        for config_key, config_value in config.items():
            config_key_type = self.__annotations__.get(config_key, None)
            if config_key_type is None or config_value is None:
                continue
            if is_dataclass(config_key_type) and isinstance(config_value, dict):
                try:
                    config[config_key] = config_key_type.load(config=config_value)
                except Exception:
                    traceback.print_exc()
                    config[config_key] = None
        return self(**config)

    @classmethod
    def parse_config(self, config: Dict = None) -> Dict:
        return config

    @classmethod
    def load_extra_config(self) -> Dict:
        return dict()

    def to_dict(self) -> Dict:
        return asdict(self)



from typing import List, Tuple

from mage_ai.orchestration.db.models.oauth import (
    Oauth2AccessToken,
    Permission,
    Role,
    User,
)
from mage_ai.settings import (
    DISABLE_NOTEBOOK_EDIT_ACCESS,
    REQUIRE_USER_AUTHENTICATION,
    is_disable_pipeline_edit_access,
)
from mage_ai.shared.environments import is_test


def authenticate_client_and_token(client_id: str, token: str) -> Tuple[Oauth2AccessToken, bool]:
    oauth_token = Oauth2AccessToken.query.filter(
        Oauth2AccessToken.oauth2_application_id == client_id,
        Oauth2AccessToken.token == token,
    ).first()

    valid = False
    if oauth_token:
        valid = oauth_token.is_valid()

    return oauth_token, valid


def is_owner(
    user: User,
    entity: Permission.Entity = None,
    entity_id: str = None,
) -> bool:
    return (user and user.owner) or \
        (not REQUIRE_USER_AUTHENTICATION and not is_test()) or \
        (user and user.get_access(entity, entity_id) & Permission.Access.OWNER != 0)


def has_at_least_admin_role(
    user: User,
    entity: Permission.Entity = None,
    entity_id: str = None,
) -> bool:
    return not user or \
        (not REQUIRE_USER_AUTHENTICATION and not is_test()) or \
        is_owner(user, entity, entity_id) or \
        (user.roles and user.roles & 1 != 0) or \
        (user and user.get_access(entity, entity_id) & Permission.Access.ADMIN != 0)


def has_at_least_editor_role(
    user: User,
    entity: Permission.Entity = None,
    entity_id: str = None,
) -> bool:
    return not user or \
        (not REQUIRE_USER_AUTHENTICATION and not is_test()) or \
        is_owner(user, entity, entity_id) or \
        has_at_least_admin_role(user, entity, entity_id) or \
        (user.roles and user.roles & 2 != 0) or \
        (user and user.get_access(entity, entity_id) & Permission.Access.EDITOR != 0)


def has_at_least_editor_role_and_notebook_edit_access(
    user: User,
    entity: Permission.Entity = None,
    entity_id: str = None,
) -> bool:
    return DISABLE_NOTEBOOK_EDIT_ACCESS != 1 and \
        has_at_least_editor_role(user, entity, entity_id)


def has_at_least_editor_role_and_pipeline_edit_access(
    user: User,
    entity: Permission.Entity = None,
    entity_id: str = None,
) -> bool:
    return not is_disable_pipeline_edit_access() and \
        has_at_least_editor_role(user, entity, entity_id)


def has_at_least_viewer_role(
    user: User,
    entity: Permission.Entity = None,
    entity_id: str = None,
) -> bool:
    return not user or \
        (not REQUIRE_USER_AUTHENTICATION and not is_test()) or \
        is_owner(user, entity, entity_id) or \
        has_at_least_admin_role(user, entity, entity_id) or \
        has_at_least_editor_role(user, entity, entity_id) or \
        (user.roles and user.roles & 4 != 0) or \
        (user and user.get_access(entity, entity_id) & Permission.Access.VIEWER != 0)


def get_access_for_roles(
    roles: List[Role],
    entity: Permission.Entity,
    entity_id: str = None,
):
    access = 0
    for role in roles:
        if role:
            access = access | role.get_access(entity, entity_id=entity_id)
    return access


def parse_cookie_header(cookies_raw):
    cookies = {}
    if cookies_raw:
        for cookie_string in cookies_raw.split(';'):
            cookie_string = cookie_string.strip()
            if "=" in cookie_string:
                cookie_name, cookie_value = cookie_string.split('=', 1)
            else:
                # Assume an empty name per
                # https://bugzilla.mozilla.org/show_bug.cgi?id=169091
                cookie_name, cookie_value = "", cookie_string
            cookies[cookie_name] = cookie_value

    return cookies



from collections import namedtuple
from dataclasses import dataclass
from typing import Callable, Dict

import pika
from mage_ai.shared.config import BaseConfig
from mage_ai.streaming.sources.base import BaseSource
from pika.exceptions import AMQPConnectionError


@dataclass
class ConsumeConfig:
    auto_ack: bool = False
    exclusive: bool = False
    inactivity_timeout: float = None


@dataclass
class RabbitMQConfig(BaseConfig):
    connection_host: str
    connection_port: int
    queue_name: str
    configure_consume: bool = False
    username: str = 'guest'
    password: str = 'guest'
    amqp_url_virtual_host: str = r'%2f'
    consume_config: ConsumeConfig = None

    @classmethod
    def parse_config(self, config: Dict) -> Dict:
        consume_config = config.get('consume_config')
        if consume_config is not None and type(consume_config) is dict:
            config['consume_config'] = ConsumeConfig(**consume_config)
        return config


class RabbitMQSource(BaseSource):
    config_class = RabbitMQConfig

    def init_client(self):
        queue_name = self.config.queue_name
        username = self.config.username
        password = self.config.password
        connection_host = self.config.connection_host
        connection_port = self.config.connection_port
        vt_host = self.config.amqp_url_virtual_host

        self._print(f'Starting to initialize consumer for queue {queue_name}')

        try:

            generated_url = f"amqp://{username}:{password}@" \
                            f"{connection_host}:{connection_port}/{vt_host}"

            self._print(f'Trying to connect on {generated_url}')
            self.create_connection = pika.BlockingConnection(
                pika.URLParameters(
                    generated_url
                )
            )

            self._print('Connected on broker')

            self.channel = self.create_connection.channel()

        except AMQPConnectionError:
            self._print('Connection Error, please check broker connection')
            raise AMQPConnectionError
        except Exception as e:
            print(e)
            raise e

    def read(self, handler: Callable):
        pass

    def batch_read(self, handler: Callable):

        self._print('Start consuming messages.')

        # using namedtuple for better usage on transformer blocks

        message_tuple = namedtuple('Payload', ['method', 'properties', 'body'])
        if self.config.configure_consume is True:

            inactivity_timeout = self.config.consume_config.inactivity_timeout

            for method, properties, body in self.channel.consume(
                    self.config.queue_name,
                    self.config.consume_config.auto_ack,
                    self.config.consume_config.exclusive,
                    arguments=None,
                    inactivity_timeout=inactivity_timeout
            ):

                full_message = message_tuple(method, properties, body)
                self.__print_message(full_message)
                handler(full_message, **{'channel': self.channel})
        else:
            for method, properties, body in self.channel.consume(
                self.config.queue_name
            ):
                full_message = message_tuple(method, properties, body)
                self.__print_message(full_message)
                handler(full_message, **{'channel': self.channel})

    def __print_message(self, method):
        self._print(f'Received message {method}')



{% extends 'data_loaders/deltalake/default.jinja' %}
{% block storage_options %}
    storage_options = {
        'AZURE_STORAGE_ACCOUNT_NAME': '',
        'AZURE_STORAGE_ACCOUNT_KEY': '',
        'AZURE_STORAGE_ACCESS_KEY': '',
        'AZURE_STORAGE_MASTER_KEY': '',
        'AZURE_STORAGE_CLIENT_ID': '',
        'AZURE_STORAGE_CLIENT_SECRET': '',
        'AZURE_STORAGE_TENANT_ID': '',
        'AZURE_STORAGE_SAS_KEY': '',
        'AZURE_STORAGE_TOKEN': '',
        'AZURE_STORAGE_USE_EMULATOR': '',
    }
{% endblock %}

{% block uri %}
    uri = 'az://[container]/[key]'
{% endblock %}



from mage_ai.services.airbyte.exceptions import (
    ConnectionNotFound,
    JobNotFound,
    UnhealthyServer,
)
from typing import Dict
import httpx
import logging


class AirbyteClient:
    def __init__(
        self,
        base_url: str = 'http://localhost:8000/api/v1',
        logger: logging.Logger = None,
        password: str = 'password',
        timeout: int = 12,
        username: str = 'airbyte',
    ):
        self.base_url = base_url
        self.logger = logger or logging.getLogger(__name__)
        self.timeout = timeout

        self._client = httpx.Client(
            auth=(username, password),
            base_url=self.base_url,
            timeout=timeout,
        )
        self._closed = False
        self._started = False

    def get_connection_status(self, connection_id: str) -> str:
        try:
            response = self._client.post(
                f'{self.base_url}/connections/get/',
                json=dict(connectionId=connection_id),
            )
            response.raise_for_status()

            return response.json()['status']
        except httpx.HTTPStatusError as e:
            if e.response.status_code == 404:
                raise ConnectionNotFound(f'Connection {connection_id} not found.') from e

            raise UnhealthyServer() from e

    def get_health_status(self, client: httpx.Client) -> bool:
        try:
            print(f'{self.base_url}/health/')
            response = self._client.get(f'{self.base_url}/health/')
            response.raise_for_status()
            data = response.json()

            self.logger.debug(f'Health check: {data}')

            key = 'available' if 'available' in data else 'db'
            status = response.json()[key]
            if not status:
                raise UnhealthyServer(f"Airbyte server health: {status}")

            return True
        except httpx.HTTPStatusError as e:
            raise UnhealthyServer() from e

    def get_job_status(self, job_id: str) -> Dict:
        try:
            response = self._client.post(
                f'{self.base_url}/jobs/get/',
                json=dict(id=job_id),
            )
            response.raise_for_status()

            return response.json()['job']
        except httpx.HTTPStatusError as e:
            if e.response.status_code == 404:
                raise JobNotFound(f'Job {job_id} not found.') from e

            raise UnhealthyServer() from e

    def run_sync(self, connection_id: str) -> Dict:
        try:
            response = self._client.post(
                f'{self.base_url}/connections/sync/',
                json=dict(connectionId=connection_id),
            )
            response.raise_for_status()

            return response.json()['job']
        except httpx.HTTPStatusError as e:
            if e.response.status_code == 404:
                raise ConnectionNotFound(f'Connection {connection_id} not found.') from e

            raise UnhealthyServer() from e

    def __enter__(self):
        if self._closed:
            raise RuntimeError('Airbyte client cannot be started after it’s been closed.')

        if self._started:
            raise RuntimeError('Airbyte client cannot be started more than once.')

        self._started = True

        self.get_health_status(self._client)

        return self

    def __exit__(self, *exc):
        self._closed = True
        self._client.__exit__()



import os

from mage_ai.api.resources.GenericResource import GenericResource
from mage_ai.data_preparation.models.constants import MAX_PRINT_OUTPUT_LINES
from mage_ai.data_preparation.repo_manager import (
    ProjectType,
    get_project_type,
    get_repo_config,
    get_repo_path,
)
from mage_ai.data_preparation.shared.constants import MANAGE_ENV_VAR
from mage_ai.orchestration.db import safe_db_query
from mage_ai.server.api.clusters import ClusterType
from mage_ai.server.scheduler_manager import scheduler_manager
from mage_ai.settings import (
    REQUIRE_USER_AUTHENTICATION,
    is_disable_pipeline_edit_access,
)


class StatusResource(GenericResource):
    """
    Resource to fetch block output for the notebook. Created to support legacy
    endpoint /api/pipelines/<pipeline_uuid>/blocks/<block_uuid>/outputs
    """
    @classmethod
    @safe_db_query
    def collection(self, query, meta, user, **kwargs):
        from mage_ai.cluster_manager.constants import (
            ECS_CLUSTER_NAME,
            GCP_PROJECT_ID,
            KUBE_NAMESPACE,
        )
        instance_type = None
        if get_project_type() == ProjectType.MAIN:
            instance_type = get_repo_config().cluster_type
        elif os.getenv(ECS_CLUSTER_NAME):
            instance_type = ClusterType.ECS
        elif os.getenv(GCP_PROJECT_ID):
            instance_type = ClusterType.CLOUD_RUN
        else:
            try:
                from mage_ai.cluster_manager.kubernetes.workload_manager import (
                    WorkloadManager,
                )
                if WorkloadManager.load_config() or os.getenv(KUBE_NAMESPACE):
                    instance_type = ClusterType.K8S
            except ModuleNotFoundError:
                pass

        status = {
            'is_instance_manager': os.getenv(MANAGE_ENV_VAR) == '1',
            'repo_path': get_repo_path(),
            'scheduler_status': scheduler_manager.get_status(),
            'instance_type': instance_type,
            'disable_pipeline_edit_access': is_disable_pipeline_edit_access(),
            'max_print_output_lines': MAX_PRINT_OUTPUT_LINES,
            'require_user_authentication': REQUIRE_USER_AUTHENTICATION,
        }
        return self.build_result_set([status], user, **kwargs)



import json
import re

import terminado
from tornado import gen

from mage_ai.api.utils import authenticate_client_and_token, has_at_least_editor_role
from mage_ai.orchestration.db.models.oauth import Oauth2Application
from mage_ai.settings import (
    REQUIRE_USER_AUTHENTICATION,
    is_disable_pipeline_edit_access,
)


class MageTermManager(terminado.NamedTermManager):
    def get_terminal(self, term_name: str, **kwargs):
        assert term_name is not None

        if term_name in self.terminals:
            return self.terminals[term_name]

        if self.max_terminals and len(self.terminals) >= self.max_terminals:
            raise terminado.management.MaxTerminalsReached(self.max_terminals)

        # Create new terminal
        self.log.info("New terminal with specified name: %s", term_name)
        term = self.new_terminal(**kwargs)
        term.term_name = term_name
        self.terminals[term_name] = term
        self.start_reading(term)
        return term


class MageUniqueTermManager(terminado.UniqueTermManager):
    def get_terminal(self, url_component=None, **kwargs):
        if self.max_terminals and len(self.ptys_by_fd) >= self.max_terminals:
            raise terminado.management.MaxTerminalsReached(self.max_terminals)

        term = self.new_terminal(**kwargs)
        self.start_reading(term)
        return term


class TerminalWebsocketServer(terminado.TermSocket):
    @property
    def term_command(self):
        return next(iter(self.term_manager.shell_command))

    def check_origin(self, origin):
        return True

    def on_pty_read(self, text):
        """Data read from pty; send to frontend"""
        updated_text = text
        if self.term_command == 'cmd':
            xterm_escape = re.compile(r'(?:\x1B\]0;).*\x07')
            updated_text = xterm_escape.sub('', text)
        self.send_json_message(["stdout", updated_text])

    def open(self, url_component=None):
        """Websocket connection opened.

        Call our terminal manager to get a terminal, and connect to it as a
        client.
        """
        # Jupyter has a mixin to ping websockets and keep connections through
        # proxies alive. Call super() to allow that to set up:
        super(terminado.TermSocket, self).open(url_component)
        api_key = self.get_argument('api_key', None, True)
        token = self.get_argument('token', None, True)

        cwd = self.get_argument('cwd', None, True)
        term_name = self.get_argument('term_name', None, True)

        user = None
        if REQUIRE_USER_AUTHENTICATION and api_key and token:
            oauth_client = Oauth2Application.query.filter(
                Oauth2Application.client_id == api_key,
            ).first()
            if oauth_client:
                oauth_token, valid = authenticate_client_and_token(oauth_client.id, token)
                valid = valid and \
                    oauth_token and \
                    oauth_token.user
                if valid:
                    user = oauth_token.user

        self.term_name = term_name if term_name else 'tty'
        if user:
            self.term_name = f'{self.term_name}_{user.id}'

        self._logger.info("TermSocket.open: %s", self.term_name)

        self.terminal = self.term_manager.get_terminal(self.term_name, cwd=cwd)
        self.terminal.clients.append(self)
        self.__initiate_terminal(self.terminal)

    @gen.coroutine
    def on_message(self, raw_message):
        message = json.loads(raw_message)

        api_key = message.get('api_key')
        token = message.get('token')
        command = message.get('command')

        if REQUIRE_USER_AUTHENTICATION or is_disable_pipeline_edit_access():
            valid = False

            if api_key and token:
                oauth_client = Oauth2Application.query.filter(
                    Oauth2Application.client_id == api_key,
                ).first()
                if oauth_client:
                    oauth_token, valid = authenticate_client_and_token(oauth_client.id, token)
                    valid = valid and \
                        oauth_token and \
                        oauth_token.user and \
                        has_at_least_editor_role(oauth_token.user)
            if not valid or is_disable_pipeline_edit_access():
                return self.send_json_message(
                    ['stdout', f'{command[1]}\nUnauthorized access to the terminal.'])

        return terminado.TermSocket.on_message(self, json.dumps(command))

    def __initiate_terminal(self, terminal):
        self.send_json_message(["setup", {}])
        self._logger.info("TermSocket.open: Opened %s", self.term_name)
        # Now drain the preopen buffer, if reconnect.
        buffered = ""
        preopen_buffer = terminal.read_buffer.copy()
        while True:
            if not preopen_buffer:
                break
            s = preopen_buffer.popleft()
            buffered += s
        if buffered:
            self.on_pty_read(buffered)

        # Turn enable-bracketed-paste off since it can mess up the output.
        if self.term_command == 'bash':
            terminal.ptyproc.write(
                "bind 'set enable-bracketed-paste off' # Mage terminal settings command\r")
        terminal.read_buffer.clear()



from great_expectations.core import ExpectationSuite
from great_expectations.core.batch import RuntimeBatchRequest
from mage_ai.data_preparation.models.constants import BlockLanguage
from typing import Any, Dict, List, Tuple
import great_expectations as gx
import pandas as pd


class GreatExpectations():
    def __init__(self, block, expectations: List[Dict] = None):
        self.block = block
        self.expectations = expectations

    def build_validators(
        self,
        *args,
        **kwargs,
    ) -> List[Tuple[Any, str]]:
        validators = []

        for idx, df in enumerate(args):
            if idx >= len(self.block.upstream_block_uuids):
                continue

            uuid = self.block.upstream_block_uuids[idx]
            expectation_suite_name = f'expectation_suite_for_block_{uuid}'

            context = gx.get_context()
            batch_request = None
            datasource_config = None

            if type(df) is list:
                df = pd.DataFrame(df)
            elif type(df) is dict:
                df = pd.DataFrame([df])

            if BlockLanguage.PYTHON == self.block.language:
                data_connector_name = f'data_connector_name_{uuid}'
                datasource_name = f'datasource_name_{uuid}'

                batch_request = RuntimeBatchRequest(
                    batch_identifiers=dict(
                        default_identifier_name='default_identifier',
                    ),
                    data_asset_name=f'data_asset_{uuid}',
                    data_connector_name=data_connector_name,
                    datasource_name=datasource_name,
                    runtime_parameters=dict(
                        batch_data=df,
                    ),
                )

                datasource_config = dict(
                    class_name='Datasource',
                    data_connectors={
                        data_connector_name: dict(
                            batch_identifiers=[
                                'default_identifier_name',
                            ],
                            class_name='RuntimeDataConnector',
                            module_name='great_expectations.datasource.data_connector',
                        ),
                    },
                    execution_engine=dict(
                        class_name='PandasExecutionEngine',
                        module_name='great_expectations.execution_engine',
                    ),
                    module_name='great_expectations.datasource',
                    name=datasource_name,
                )

            elif BlockLanguage.SQL == self.block.language:
                pass

            context.add_datasource(**datasource_config)

            validator_options = dict(batch_request=batch_request)

            if self.expectations and len(self.expectations) >= 1:
                validator_options['expectation_suite'] = ExpectationSuite(
                    expectation_suite_name=expectation_suite_name,
                    expectations=self.expectations,
                )
            else:
                context.add_or_update_expectation_suite(
                    expectation_suite_name=expectation_suite_name,
                )
                validator_options['expectation_suite_name'] = expectation_suite_name

            validator = context.get_validator(**validator_options)

            validators.append((validator, uuid))

        return validators



import os
import shutil
from typing import Dict, List

import yaml

from mage_ai.api.errors import ApiError
from mage_ai.api.resources.GenericResource import GenericResource
from mage_ai.cluster_manager.constants import (
    ECS_CLUSTER_NAME,
    ECS_CONTAINER_NAME,
    ECS_TASK_DEFINITION,
    GCP_PATH_TO_KEYFILE,
    GCP_PROJECT_ID,
    GCP_REGION,
    KUBE_NAMESPACE,
    KUBE_SERVICE_ACCOUNT_NAME,
    KUBE_STORAGE_CLASS_NAME,
)
from mage_ai.data_preparation.repo_manager import (
    ProjectType,
    get_project_type,
    get_repo_config,
    get_repo_path,
    init_repo,
)
from mage_ai.data_preparation.shared.constants import MANAGE_ENV_VAR
from mage_ai.orchestration.db import safe_db_query
from mage_ai.orchestration.db.models.oauth import Permission, Role, User
from mage_ai.server.api.clusters import ClusterType
from mage_ai.server.logger import Logger

logger = Logger().new_server_logger(__name__)


class WorkspaceResource(GenericResource):
    @classmethod
    @safe_db_query
    def collection(self, query_arg, meta, user, **kwargs):
        cluster_type = self.verify_project()
        if not cluster_type:
            cluster_type = query_arg.get('cluster_type', [None])
            if cluster_type:
                cluster_type = cluster_type[0]

        user_id = query_arg.get('user_id', [None])
        query_user = None
        if user_id:
            user_id = user_id[0]
            query_user = User.query.get(user_id)

        instances = self.get_instances(cluster_type)
        instance_map = {
            instance.get('name'): instance
            for instance in instances
        }

        is_main_project = get_project_type() == ProjectType.MAIN

        repo_path = get_repo_path()
        projects_folder = os.path.join(repo_path, 'projects')
        if is_main_project:
            projects = [f.name for f in os.scandir(projects_folder) if f.is_dir()]
        else:
            projects = [instance.get('name') for instance in instances]

        workspaces = []
        for project in projects:
            if project in instance_map:
                repo_path = os.path.join(projects_folder, project)
                workspace = dict(
                    name=project,
                    repo_path=repo_path,
                    cluster_type=cluster_type,
                    instance=instance_map[project],
                )
                if is_main_project and query_user:
                    workspace['access'] = query_user.get_access(
                        Permission.Entity.PROJECT,
                        repo_path,
                    )
                workspaces.append(workspace)

        return self.build_result_set(workspaces, user, **kwargs)

    @classmethod
    @safe_db_query
    def member(self, pk, user, **kwargs):
        cluster_type = self.verify_project(pk)
        if not cluster_type:
            query = kwargs.get('query', {})
            cluster_type = query.get('cluster_type')[0]

        instances = self.get_instances(cluster_type)
        instance_map = {
            instance.get('name'): instance
            for instance in instances
        }

        return self(dict(
            name=pk,
            cluster_type=cluster_type,
            instance=instance_map[pk],
        ), user, **kwargs)

    @classmethod
    @safe_db_query
    def create(self, payload, user, **kwargs):
        cluster_type = self.verify_project()
        if not cluster_type:
            cluster_type = payload.get('cluster_type')

        workspace_name = payload.get('name')

        error = ApiError.RESOURCE_ERROR.copy()

        workspace_folder = None
        if get_project_type() == ProjectType.MAIN:
            workspace_folder = os.path.join(get_repo_path(), 'projects', workspace_name)
            if os.path.exists(workspace_folder):
                error.update(message=f'Project with name {workspace_name} already exists')
                raise ApiError(error)
            try:
                init_repo(workspace_folder, project_type=ProjectType.SUB)
            except Exception as e:
                error.update(message=f'Error creating project: {str(e)}')
                raise ApiError(error)
        try:
            if cluster_type == ClusterType.K8S:
                from mage_ai.cluster_manager.kubernetes.workload_manager import (
                    WorkloadManager,
                )
                namespace = payload.get('namespace', os.getenv(KUBE_NAMESPACE))
                storage_class_name = payload.get(
                    'storage_class_name',
                    os.getenv(KUBE_STORAGE_CLASS_NAME),
                )
                service_account_name = payload.get(
                    'service_account_name',
                    os.getenv(KUBE_SERVICE_ACCOUNT_NAME),
                )
                container_config_yaml = payload.get('container_config')
                container_config = None
                if container_config_yaml:
                    container_config = yaml.full_load(container_config_yaml)

                k8s_workload_manager = WorkloadManager(namespace)
                k8s_workload_manager.create_deployment(
                    workspace_name,
                    container_config=container_config,
                    service_account_name=service_account_name,
                    storage_class_name=storage_class_name,
                    volume_host_path=workspace_folder,
                )
            elif cluster_type == ClusterType.ECS:
                from mage_ai.cluster_manager.aws.ecs_task_manager import EcsTaskManager
                cluster_name = payload.get('cluster_name', os.getenv(ECS_CLUSTER_NAME))
                task_definition = payload.get(
                    'task_definition',
                    os.getenv(ECS_TASK_DEFINITION),
                )
                container_name = payload.get(
                    'container_name',
                    os.getenv(ECS_CONTAINER_NAME),
                )

                ecs_instance_manager = EcsTaskManager(cluster_name)

                ecs_instance_manager.create_task(
                    workspace_name,
                    task_definition,
                    container_name,
                )
            elif cluster_type == ClusterType.CLOUD_RUN:
                from mage_ai.cluster_manager.gcp.cloud_run_service_manager import (
                    CloudRunServiceManager,
                )
                project_id = payload.get('project_id', os.getenv(GCP_PROJECT_ID))
                path_to_credentials = payload.get(
                    'path_to_credentials',
                    os.getenv('path_to_keyfile')
                )
                region = payload.get('region', os.getenv('GCP_REGION'))

                cloud_run_service_manager = CloudRunServiceManager(
                    project_id,
                    path_to_credentials,
                    region=region
                )

                cloud_run_service_manager.create_service(workspace_name)
        except Exception as e:
            if workspace_folder and os.path.exists(workspace_folder):
                shutil.rmtree(workspace_folder)
            return self(dict(success=False, error=str(e)), user, **kwargs)

        if get_project_type() == ProjectType.MAIN:
            Role.create_default_roles(
                entity=Permission.Entity.PROJECT,
                entity_id=workspace_folder,
                prefix=workspace_name,
            )

        return self(dict(success=True), user, **kwargs)

    def update(self, payload, **kwargs):
        cluster_type = self.model.get('cluster_type')
        instance_name = self.model.get('name')
        if cluster_type == 'ecs':
            from mage_ai.cluster_manager.aws.ecs_task_manager import EcsTaskManager
            task_arn = payload.get('task_arn')
            cluster_name = payload.get('cluster_name', os.getenv(ECS_CLUSTER_NAME))
            task_definition = payload.get(
                'task_definition',
                os.getenv(ECS_TASK_DEFINITION),
            )
            container_name = payload.get('container_name', os.getenv(ECS_CONTAINER_NAME))

            action = payload.get('action')

            ecs_instance_manager = EcsTaskManager(cluster_name)
            if action == 'stop':
                ecs_instance_manager.stop_task(task_arn)
            elif action == 'resume':
                ecs_instance_manager.create_task(
                    instance_name,
                    task_definition,
                    container_name,
                )

        return self

    def delete(self, **kwargs):
        cluster_type = self.model.get('cluster_type')
        workspace_name = self.model.get('name')
        instance = self.model.get('instance')

        repo_path = get_repo_path()
        project_folder = os.path.join(repo_path, 'projects', workspace_name)

        if get_project_type() == ProjectType.MAIN:
            shutil.rmtree(project_folder)
        if cluster_type == 'ecs':
            from mage_ai.cluster_manager.aws.ecs_task_manager import EcsTaskManager
            task_arn = instance.get('task_arn')
            cluster_name = os.getenv(ECS_CLUSTER_NAME)

            ecs_instance_manager = EcsTaskManager(cluster_name)
            ecs_instance_manager.delete_task(workspace_name, task_arn)

        return self

    @classmethod
    def verify_project(self, subproject: str = None) -> str:
        project_type = get_project_type()
        if project_type != ProjectType.MAIN and os.getenv(MANAGE_ENV_VAR) != '1':
            error = ApiError.RESOURCE_ERROR.copy()
            error.update(message='This project is ineligible for workspace management.')
            raise ApiError(error)

        if project_type == ProjectType.MAIN and subproject:
            repo_path = get_repo_path()
            projects_folder = os.path.join(repo_path, 'projects')
            projects = [name for name in os.listdir(projects_folder) if os.path.isdir(name)]
            if subproject not in projects:
                error = ApiError.RESOURCE_NOT_FOUND.copy()
                error.update(message=f'Project {subproject} was not found.')
                raise ApiError(error)

        if project_type == ProjectType.MAIN:
            return get_repo_config().cluster_type

    @classmethod
    def get_instances(self, cluster_type: str) -> List[Dict]:
        instances = []
        if cluster_type == ClusterType.K8S:
            from mage_ai.cluster_manager.kubernetes.workload_manager import (
                WorkloadManager,
            )
            namespace = os.getenv(KUBE_NAMESPACE)
            workload_manager = WorkloadManager(namespace)

            instances = workload_manager.list_services()
        elif cluster_type == ClusterType.ECS:
            from mage_ai.cluster_manager.aws.ecs_task_manager import EcsTaskManager

            try:
                cluster_name = os.getenv(ECS_CLUSTER_NAME)
                ecs_instance_manager = EcsTaskManager(cluster_name)
                instances = ecs_instance_manager.list_tasks()
            except Exception as e:
                error = ApiError.RESOURCE_ERROR.copy()
                error.update(message=str(e))
                raise ApiError(error)
        elif cluster_type == ClusterType.CLOUD_RUN:
            from mage_ai.cluster_manager.gcp.cloud_run_service_manager import (
                CloudRunServiceManager,
            )
            project_id = os.getenv(GCP_PROJECT_ID)
            path_to_credentials = os.getenv(GCP_PATH_TO_KEYFILE)
            region = os.getenv(GCP_REGION)
            cloud_run_service_manager = CloudRunServiceManager(
                project_id,
                path_to_credentials,
                region=region
            )

            instances = cloud_run_service_manager.list_services()

        return instances



LOG_LEVEL_DEBUG = 'DEBUG'
LOG_LEVEL_ERROR = 'ERROR'
LOG_LEVEL_EXCEPTION = 'EXCEPTION'
LOG_LEVEL_INFO = 'INFO'

TYPE_ACTIVATE_VERSION = 'ACTIVATE_VERSION'
TYPE_LOG = 'LOG'
TYPE_RECORD = 'RECORD'
TYPE_SAMPLE_DATA = 'SAMPLE_DATA'
TYPE_SCHEMA = 'SCHEMA'
TYPE_STATE = 'STATE'



import pandas as pd


def format_value(value):
    if value is None or pd.isnull(value):
        return 'NULL'

    if type(value) is bool:
        if value is True:
            return 'TRUE'
        else:
            return 'FALSE'

    if type(value) is int or type(value) is float:
        return str(value)

    if type(value) is str:
        value = escape_quotes(value)

    return f"'{value}'"


def escape_quotes(line: str, single: bool = True, double: bool = True) -> str:
    new_line = str(line)
    if single:
        new_line = new_line.replace("'", "''")
    if double:
        new_line = new_line.replace('\"', '\\"')
    return new_line



from mage_ai.orchestration.db.models.oauth import User
from mage_ai.shared.array import find
import importlib


def build_load(associated_resource_class, **kwargs):
    associated_id_column_name = '{}_id'.format(kwargs.get(
        'attribute', associated_resource_class.resource_name_singular()), )

    def _load(resource):
        ids = list(
            filter(
                lambda x: x, [
                    getattr(
                        r, associated_id_column_name) for r in resource.result_set()], ))

        if ids:
            query_set = associated_resource_class.model_class.objects.filter(
                id__in=ids)
        else:
            query_set = []

        return getattr(
            importlib.import_module(
                'api.resources.{}'.format(
                    associated_resource_class.__name__)),
            associated_resource_class.__name__,
        ).build_result_set(
            query_set,
            resource.current_user)

    return _load


def build_load_select(current_resource, associated_resource_class, **kwargs):
    associated_id_column_name = '{}_id'.format(
        kwargs.get('attribute', current_resource.resource_name_singular()),
    )

    def _load(resource):
        ids = [r.id for r in resource.result_set()]
        query = {}
        query[f'{associated_id_column_name}__in'] = ids
        query_set = associated_resource_class.model_class.objects.filter(
            **query)
        return getattr(
            importlib.import_module(
                'api.resources.{}'.format(
                    associated_resource_class.__name__)),
            associated_resource_class.__name__,
        ).build_result_set(
            query_set,
            resource.current_user)
    return _load


def build_select_find(query_column_name):
    def _select(resource, arr):
        return find(
            lambda res: res.id == getattr(
                resource,
                query_column_name),
            arr)
    return _select


def build_select_filter(query_column_name):
    def _select(resource, arr):
        return list(
            filter(
                lambda res: resource.id == getattr(
                    res,
                    query_column_name),
                arr))
    return _select


def load_user(resource):
    user_ids = [r.user_id for r in resource.result_set()]
    query_set = User.query.filter(User.id.in_(user_ids))
    UserResource = getattr(
        importlib.import_module('mage_ai.api.resources.UserResource'),
        'UserResource',
    )
    return UserResource.build_result_set(query_set, resource.current_user)


def select_user(resource, arr):
    return find(lambda res: res.id == resource.user_id, arr)



from mage_ai.api.resources.BaseResource import BaseResource


class GenericResource(BaseResource):
    def __getattr__(self, name):
        def _missing(*args, **kwargs):
            if type(self.model) is dict:
                return self.model.get(name)
            return self.model
        return _missing()



import os
from typing import Dict

from mage_ai.orchestration.notification.config import (
    AlertOn,
    MessageTemplate,
    NotificationConfig,
)
from mage_ai.services.email.email import send_email
from mage_ai.services.google_chat.google_chat import send_google_chat_message
from mage_ai.services.opsgenie.opsgenie import send_opsgenie_alert
from mage_ai.services.slack.slack import send_slack_message
from mage_ai.services.teams.teams import send_teams_message
from mage_ai.settings import DEFAULT_LOCALHOST_URL, MAGE_PUBLIC_HOST

DEFAULT_MESSAGES = dict(
    success=dict(
        title='Successfully ran Pipeline {pipeline_uuid}',
        summary=(
            'Successfully ran Pipeline `{pipeline_uuid}` with Trigger {pipeline_schedule_id} '
            '`{pipeline_schedule_name}` at execution time `{execution_time}`.'
        ),
    ),
    failure=dict(
        title='Failed to run Mage pipeline {pipeline_uuid}',
        summary=(
            'Failed to run Pipeline `{pipeline_uuid}` with Trigger {pipeline_schedule_id} '
            '`{pipeline_schedule_name}` at execution time `{execution_time}`.'
        ),
    ),
    passed_sla=dict(
        title='SLA passed for Mage pipeline {{pipeline_uuid}}',
        summary=(
            'SLA passed for pipeline `{pipeline_uuid}` with Trigger {pipeline_schedule_id} '
            '`{pipeline_schedule_name}` at execution time `{execution_time}`.'
        ),
    ),
)


class NotificationSender:
    def __init__(self, config: NotificationConfig):
        self.config = config

    def send(
        self,
        title: str = None,
        summary: str = None,
        details: str = None,
    ) -> None:
        """Send messages to the notification channels.

        Args:
            title (str, optional): Short sentence, used as title (e.g. Email subject)
            summary (str, optional): Mid-length sentences, used as the summary of the message.
            details (str, optional): Long message, used as the body of the message (e.g. Email body)
        """
        if summary is None:
            return
        if self.config.slack_config is not None and self.config.slack_config.is_valid:
            send_slack_message(self.config.slack_config, details or summary)

        if self.config.teams_config is not None and self.config.teams_config.is_valid:
            send_teams_message(self.config.teams_config, summary)

        if self.config.google_chat_config is not None and self.config.google_chat_config.is_valid:
            send_google_chat_message(self.config.google_chat_config, summary)

        if self.config.email_config is not None and title is not None:
            send_email(
                self.config.email_config,
                subject=title,
                message=details or summary,
            )

        if self.config.opsgenie_config is not None and self.config.opsgenie_config.is_valid:
            send_opsgenie_alert(
                self.config.opsgenie_config,
                message=title,
                description=details or summary,
            )

    def send_pipeline_run_success_message(self, pipeline, pipeline_run) -> None:
        if AlertOn.PIPELINE_RUN_SUCCESS in self.config.alert_on:
            default_message = DEFAULT_MESSAGES['success']
            if self.config.message_templates:
                message_template = self.config.message_templates.success
            else:
                message_template = None
            self.__send_pipeline_run_message(
                default_message,
                pipeline,
                pipeline_run,
                message_template=message_template,
            )

    def send_pipeline_run_failure_message(
        self,
        pipeline,
        pipeline_run,
        summary: str = None,
    ) -> None:
        if AlertOn.PIPELINE_RUN_FAILURE in self.config.alert_on:
            default_message = DEFAULT_MESSAGES['failure']
            if self.config.message_templates:
                message_template = self.config.message_templates.failure
            else:
                message_template = None
            self.__send_pipeline_run_message(
                default_message,
                pipeline,
                pipeline_run,
                message_template=message_template,
                summary=summary,
            )

    def send_pipeline_run_sla_passed_message(self, pipeline, pipeline_run) -> None:
        if AlertOn.PIPELINE_RUN_PASSED_SLA in self.config.alert_on:
            default_message = DEFAULT_MESSAGES['passed_sla']
            if self.config.message_templates:
                message_template = self.config.message_templates.passed_sla
            else:
                message_template = None
            self.__send_pipeline_run_message(
                default_message,
                pipeline,
                pipeline_run,
                message_template=message_template,
            )

    def __interpolate_vars(self, text: str, pipeline, pipeline_run):
        if text is None or pipeline is None or pipeline_run is None:
            return text
        return text.format(
            execution_time=pipeline_run.execution_date,
            pipeline_run_url=self.__pipeline_run_url(pipeline, pipeline_run),
            pipeline_schedule_id=pipeline_run.pipeline_schedule.id,
            pipeline_schedule_name=pipeline_run.pipeline_schedule.name,
            pipeline_uuid=pipeline.uuid,
        )

    def __send_pipeline_run_message(
        self,
        default_message: Dict,
        pipeline,
        pipeline_run,
        message_template: MessageTemplate = None,
        summary: str = None,
    ):
        """Shared method to send pipeline run message of multiple types (success, failure, etc.).
        Priority of constructing message payload.
        1. If `summary` is provided in the method kwargs, use the `summary`
        2. If `message_template` is not None, use the user defined `message_template`.
        3. If any of `title`, `summary`, `details` is None after the steps above, get the value from
            `default_message`.

        Args:
            default_message (TYPE): default message dict, containing "title",
                "summary", "details" keys.
            pipeline: the pipeline object, used to interpolate variables in the message.
            pipeline_run: the pipeline run object, used to interpolate variables in the message.
            message_template (MessageTemplate, optional): custom message template that's provided
                by user.
            summary (str, optional): summary that's used to override the custom message template.
        """
        title = None
        details = None

        if summary is not None:
            details = self.__with_pipeline_run_url(summary, pipeline, pipeline_run)

        default_title = default_message['title']
        default_summary = default_message['summary']
        default_details = self.__with_pipeline_run_url(default_summary, pipeline, pipeline_run)

        if message_template is not None:
            if title is None and message_template.title is not None:
                title = message_template.title
            if summary is None and message_template.summary is not None:
                summary = message_template.summary
            if details is None and message_template.details is not None:
                details = message_template.details

        self.send(
            title=self.__interpolate_vars(title or default_title, pipeline, pipeline_run),
            summary=self.__interpolate_vars(summary or default_summary, pipeline, pipeline_run),
            details=self.__interpolate_vars(details or default_details, pipeline, pipeline_run),
        )

    def __with_pipeline_run_url(self, text, pipeline, pipeline_run):
        if text is None:
            return text
        text = f'{text}\n'
        if os.getenv('ENV') != 'production' or MAGE_PUBLIC_HOST != DEFAULT_LOCALHOST_URL:
            """
            Include the URL for the following cases
            1. Dev environment: Use the default localhost as host in URL
            2. Production environment: If MAGE_PUBLIC_HOST is set, use it as host.
            """
            text += f'Open {self.__pipeline_run_url(pipeline, pipeline_run)} '\
                    'to check pipeline run results and logs.'
        return text

    @staticmethod
    def __pipeline_run_url(pipeline, pipeline_run):
        return f'{MAGE_PUBLIC_HOST}/pipelines/{pipeline.uuid}/runs/'\
               f'{pipeline_run.id}'



from mage_ai.orchestration.db.models.oauth import Role, User
from mage_ai.tests.api.operations.test_base import BaseApiTestCase
from mage_ai.tests.factory import create_user


class UserOperationTests(BaseApiTestCase):
    model_class = User

    @classmethod
    def setUpClass(self):
        super().setUpClass()
        Role.create_default_roles()

    async def test_execute_create(self):
        email = self.faker.email()
        response = await self.base_test_execute_create(dict(
            email=email,
            password='water_lightning',
            password_confirmation='water_lightning',
            roles_new=[Role.get_role('Owner').id],
        ))
        self.assertEqual(User.query.get(response['user']['id']).email, email)

    async def test_execute_create_unauthorized(self):
        async def _func():
            await self.base_test_execute_create(dict(
                email=self.faker.email(),
                password='water_lightning',
                password_confirmation='water_lightning',
            ), user=create_user(), after_create_count=3, before_create_count=2)

        await self.assertRaisesAsync(Exception, _func)

    async def test_execute_delete(self):
        user = create_user(email=self.faker.email())

        await self.base_test_execute_delete(user.id)

    async def test_execute_delete_unauthorized(self):
        user = create_user(email=self.faker.email())

        async def _func():
            await self.base_test_execute_delete(user.id, user=user)

        await self.assertRaisesAsync(Exception, _func)

    async def test_execute_detail(self):
        user = create_user()
        await self.base_test_execute_detail(user.id, dict(
            email=user.email,
            username=user.username,
        ))

    async def test_execute_detail_unauthorized(self):
        user = create_user()

        async def _func():
            await self.base_test_execute_detail(user.id, dict(
                email=user.email,
                username=user.username,
            ), user=create_user())

        await self.assertRaisesAsync(Exception, _func)

    async def test_execute_list(self):
        owner = create_user(_owner=True)

        email1 = self.faker.email()
        email2 = self.faker.email()

        await self.base_test_execute_list(
            [
                dict(
                    email=email1,
                    password='water_lightning',
                    password_confirmation='water_lightning',
                    roles_new=[Role.get_role('Editor').id],
                ),
                dict(
                    email=email2,
                    password='water_lightning',
                    password_confirmation='water_lightning',
                    roles_new=[Role.get_role('Editor').id],
                ),
            ],
            [
                'id',
            ],
            user=owner,
        )

    async def test_execute_list_unauthorized(self):
        async def _func():
            await self.base_test_execute_list([], user=create_user())

        await self.assertRaisesAsync(Exception, _func)

    async def test_execute_update(self):
        user = create_user()
        new_username = self.faker.name()

        await self.base_test_execute_update(
            user.id,
            dict(username=new_username),
            dict(username=new_username),
            user=user,
        )

    async def test_execute_update_unauthorized(self):
        user = create_user()

        async def _func():
            await self.base_test_execute_update(user.id, {}, {}, user=create_user())

        await self.assertRaisesAsync(Exception, _func)



import enum


API_ENDPOINT = 'https://api.mage.ai/v1/usage_statistics'


class EventActionType(str, enum.Enum):
    IMPRESSION = 'impression'


class EventObjectType(str, enum.Enum):
    ERROR = 'error'
    PIPELINE = 'pipeline'
    PIPELINE_RUN = 'pipeline_run'
    PROJECT = 'project'
    USER = 'user'



import asyncio
import copy
import logging
import os
from contextlib import redirect_stderr, redirect_stdout
from typing import Callable, Dict, List, Union

import yaml
from jinja2 import Template

from mage_ai.data_preparation.executors.pipeline_executor import PipelineExecutor
from mage_ai.data_preparation.logging.logger import DictLogger
from mage_ai.data_preparation.models.constants import BlockType
from mage_ai.data_preparation.models.pipeline import Pipeline
from mage_ai.data_preparation.shared.stream import StreamToLogger
from mage_ai.data_preparation.shared.utils import get_template_vars
from mage_ai.shared.hash import merge_dict


class StreamingPipelineExecutor(PipelineExecutor):
    def __init__(self, pipeline: Pipeline, **kwargs):
        super().__init__(pipeline, **kwargs)
        # TODO: Support custom log destination for streaming pipelines
        self.parse_and_validate_blocks()

    def parse_and_validate_blocks(self):
        """
        Find the first valid streaming pipeline is in the structure:
        source -> transformer1 -> sink2
               -> transformer2 -> sink2
               -> transformer3 -> sink3
        """
        blocks = self.pipeline.blocks_by_uuid.values()
        source_blocks = []
        sink_blocks = []
        transformer_blocks = []
        for b in blocks:
            if b.type == BlockType.DATA_LOADER:
                # Data loader block should be root block
                if len(b.upstream_blocks or []) > 0:
                    raise Exception(f'Data loader {b.uuid} can\'t have upstream blocks.')
                if len(b.downstream_blocks or []) < 1:
                    raise Exception(f'Data loader {b.uuid} must have at least one transformer or'
                                    ' data exporter as the downstream block.')
                source_blocks.append(b)
            if b.type == BlockType.DATA_EXPORTER:
                # Data exporter block should be leaf block
                if len(b.downstream_blocks or []) > 0:
                    raise Exception(f'Data expoter {b.uuid} can\'t have downstream blocks.')
                if len(b.upstream_blocks or []) != 1:
                    raise Exception(f'Data exporter {b.uuid} must have a transformer or data'
                                    ' loader as the upstream block.')
                sink_blocks.append(b)
            if b.type == BlockType.TRANSFORMER:
                # Each transformer block can only have one upstream block
                if len(b.upstream_blocks or []) != 1:
                    raise Exception(f'Transformer {b.uuid} should (only) have one upstream block.')
                transformer_blocks.append(b)

        if len(source_blocks) != 1:
            raise Exception('Please provide (only) one data loader block as the source.')

        self.source_block = source_blocks[0]
        self.sink_blocks = sink_blocks

    def execute(
        self,
        build_block_output_stdout: Callable[..., object] = None,
        global_vars: Dict = None,
        **kwargs,
    ) -> None:
        # TODOs:
        # 1. Support multiple sources and sinks
        # 2. Support flink pipeline

        tags = self._build_tags(**kwargs)
        if build_block_output_stdout:
            stdout_logger = logging.getLogger('streaming_pipeline_executor')
            self.logger = DictLogger(stdout_logger)
            stdout = build_block_output_stdout(self.pipeline.uuid)
        else:
            self.logger = DictLogger(self.logger_manager.logger, logging_tags=tags)
            stdout = StreamToLogger(self.logger, logging_tags=tags)
        try:
            with redirect_stdout(stdout):
                with redirect_stderr(stdout):
                    self.__execute_in_python(
                        build_block_output_stdout=build_block_output_stdout,
                    )
        except Exception as e:
            if not build_block_output_stdout:
                self.logger.exception(
                        f'Failed to execute streaming pipeline {self.pipeline.uuid}',
                        **merge_dict(dict(error=e), tags),
                    )
            raise e

    def __execute_in_python(self, build_block_output_stdout: Callable[..., object] = None):
        from mage_ai.streaming.sinks.sink_factory import SinkFactory
        from mage_ai.streaming.sources.base import SourceConsumeMethod
        from mage_ai.streaming.sources.source_factory import SourceFactory
        source_config = self.__interpolate_vars(self.source_block.content)

        source = SourceFactory.get_source(
            source_config,
            checkpoint_path=os.path.join(
                self.pipeline.pipeline_variables_dir,
                'streaming_checkpoint',
            ),
        )

        sinks_by_uuid = dict()
        for sink_block in self.sink_blocks:
            sinks_by_uuid[sink_block.uuid] = SinkFactory.get_sink(
                self.__interpolate_vars(sink_block.content),
                buffer_path=os.path.join(
                    self.pipeline.pipeline_variables_dir,
                    'buffer',
                ),
            )

        def handle_batch_events_recursively(curr_block, outputs_by_block: Dict, **kwargs):
            curr_block_output = outputs_by_block[curr_block.uuid]
            for downstream_block in curr_block.downstream_blocks:
                if downstream_block.type == BlockType.TRANSFORMER:
                    execute_block_kwargs = dict(
                        global_vars=kwargs,
                        input_args=[copy.deepcopy(curr_block_output)],
                        logger=self.logger,
                    )
                    if build_block_output_stdout:
                        execute_block_kwargs['build_block_output_stdout'] = \
                            build_block_output_stdout
                    outputs_by_block[downstream_block.uuid] = \
                        downstream_block.execute_block(
                            **execute_block_kwargs,
                    )['output']
                elif downstream_block.type == BlockType.DATA_EXPORTER:
                    sinks_by_uuid[downstream_block.uuid].batch_write(
                        copy.deepcopy(curr_block_output))
                if downstream_block.downstream_blocks:
                    handle_batch_events_recursively(
                        downstream_block,
                        outputs_by_block,
                        **kwargs,
                    )

        def handle_batch_events(messages: List[Union[Dict, str]], **kwargs):
            # Handle the events with DFS

            outputs_by_block = dict()
            outputs_by_block[self.source_block.uuid] = messages

            handle_batch_events_recursively(self.source_block, outputs_by_block, **kwargs)

        async def handle_event_async(message, **kwargs):
            outputs_by_block = dict()
            outputs_by_block[self.source_block.uuid] = [message]

            handle_batch_events_recursively(self.source_block, outputs_by_block, **kwargs)

        # Long running method
        if source.consume_method == SourceConsumeMethod.BATCH_READ:
            source.batch_read(handler=handle_batch_events)
        elif source.consume_method == SourceConsumeMethod.READ_ASYNC:
            loop = asyncio.get_event_loop()
            if loop is not None:
                loop.run_until_complete(source.read_async(handler=handle_event_async))
            else:
                asyncio.run(source.read_async(handler=handle_event_async))

    def __excute_in_flink(self):
        """
        TODO: Implement this method
        """
        pass

    def __interpolate_vars(self, content):
        config_file = Template(content).render(
            **get_template_vars()
        )
        return yaml.safe_load(config_file)



from mage_ai.services.airbyte.client import AirbyteClient
from mage_ai.services.airbyte.config import AirbyteConfig
import logging


class AirbyteServer():
    def __init__(self, config: AirbyteConfig):
        self.config = config

    @property
    def protocol(self) -> str:
        return 'https' if self.config.use_ssl else 'http'

    @property
    def base_url(self) -> str:
        base_path = f'{self.config.host}:{self.config.port}'

        return f'{self.protocol}://{base_path}/api/{self.config.api_version}'

    def get_client(self, logger: logging.Logger, timeout: int = 12) -> AirbyteClient:
        return AirbyteClient(
            base_url=self.base_url,
            logger=logger,
            password=self.config.password,
            timeout=timeout,
            username=self.config.username,
        )



import json
import os
from typing import Dict

from google.api.launch_stage_pb2 import LaunchStage
from google.api_core.exceptions import AlreadyExists
from google.cloud import run_v2
from google.oauth2 import service_account

from mage_ai.server.logger import Logger
from mage_ai.services.gcp.cloud_run.config import CloudRunConfig

logger = Logger().new_server_logger(__name__)


CLOUD_RUN_CLIENT_SCOPES = [
    'https://www.googleapis.com/auth/cloud-platform',
    'https://www.googleapis.com/auth/compute',
]


def run_job(command: str, job_id: str, cloud_run_config: CloudRunConfig) -> Dict:
    if type(cloud_run_config) is dict:
        cloud_run_config = CloudRunConfig.load(config=cloud_run_config)

    if cloud_run_config.path_to_credentials_json_file:
        credentials = service_account.Credentials.from_service_account_file(
            cloud_run_config.path_to_credentials_json_file,
            scopes=CLOUD_RUN_CLIENT_SCOPES,
        )
    else:
        credentials = None
    jobs_client = run_v2.JobsClient(credentials=credentials)
    services_client = run_v2.ServicesClient(credentials=credentials)

    # Get existing service
    resource_prefix = f'projects/{cloud_run_config.project_id}/locations/{cloud_run_config.region}'
    existing_service_name = os.getenv('GCP_SERVICE_NAME')
    existing_service = services_client.get_service(
        run_v2.GetServiceRequest(
            name=f'{resource_prefix}/services/{existing_service_name}'
        )
    )
    service_template = existing_service.template

    # Create job
    containers_with_cmd = service_template.containers
    for c in containers_with_cmd:
        c.command = command.split(' ')
        c.startup_probe = None
    execution_template = run_v2.ExecutionTemplate(
        task_count=1,
        template=run_v2.TaskTemplate(
            containers=containers_with_cmd,
            volumes=service_template.volumes,
            service_account=service_template.service_account,
            execution_environment=service_template.execution_environment,
            encryption_key=service_template.encryption_key,
            vpc_access=service_template.vpc_access
        )
    )
    job = run_v2.Job(
        launch_stage=LaunchStage.BETA,
        template=execution_template,
    )

    request = run_v2.CreateJobRequest(
        parent=resource_prefix,
        job=job,
        job_id=job_id,
    )
    try:
        operation = jobs_client.create_job(request=request)
        logger.info("Waiting for create_job operation to complete...")
        response = operation.result()
        logger.info(json.dumps(response, indent=4, default=str))
    except AlreadyExists:
        pass

    job_name = f'{resource_prefix}/jobs/{job_id}'

    # Run job
    operation = jobs_client.run_job(request=run_v2.RunJobRequest(
        name=job_name,
    ))
    logger.info('Waiting for run_job operation to complete...')
    response = operation.result()
    logger.info(json.dumps(response, indent=4, default=str))

    # Delete the job after job completes
    delete_request = run_v2.DeleteJobRequest(
        name=job_name,
    )
    delete_operation = jobs_client.delete_job(request=delete_request)
    delete_response = delete_operation.result()
    logger.info('Waiting for delete_job operation to complete...')
    logger.info(json.dumps(delete_response, indent=4, default=str))

    return response



from mage_ai.data_preparation.logging import LoggerType
from mage_ai.data_preparation.logging.logger_manager import LoggerManager
from mage_ai.data_preparation.repo_manager import RepoConfig


class LoggerManagerFactory:
    @classmethod
    def get_logger_manager(
        self,
        repo_config: RepoConfig = None,
        **kwargs,
    ):
        if repo_config is not None and repo_config.logging_config is not None:
            logger_type = repo_config.logging_config.get('type')
            if logger_type == LoggerType.S3:
                from mage_ai.data_preparation.logging.s3_logger_manager import S3LoggerManager
                return S3LoggerManager(repo_config=repo_config, **kwargs)
            elif logger_type == LoggerType.GCS:
                from mage_ai.data_preparation.logging.gcs_logger_manager import GCSLoggerManager
                return GCSLoggerManager(repo_config=repo_config, **kwargs)

        return LoggerManager(repo_config=repo_config, **kwargs)



from mage_ai.api.resources.GenericResource import GenericResource
from mage_ai.data_integrations.destinations.constants import DESTINATIONS
from mage_ai.data_preparation.models.constants import BlockType
from mage_ai.data_preparation.models.pipelines.integration_pipeline import IntegrationPipeline
from mage_ai.orchestration.db import safe_db_query
from mage_ai.server.api.integration_sources import get_collection


class IntegrationDestinationResource(GenericResource):
    @classmethod
    @safe_db_query
    async def collection(self, query, meta, user, **kwargs):
        collection = get_collection('destinations', DESTINATIONS)

        return self.build_result_set(
            collection,
            user,
            **kwargs,
        )

    @classmethod
    @safe_db_query
    def create(self, payload, user, **kwargs):
        error_message = None
        success = False

        action_type = payload['action_type']
        if 'test_connection' == action_type:
            pipeline_uuid = payload['pipeline_uuid']
            pipeline = IntegrationPipeline.get(pipeline_uuid)
            config = payload['config']

            try:
                pipeline.test_connection(BlockType.DATA_EXPORTER, config=config)
                success = True
            except Exception as e:
                error_message = str(e)

        return self(dict(error_message=error_message, success=success), user, **kwargs)



from mage_ai.data_preparation.models.constants import BlockLanguage, BlockType
from mage_ai.shared.hash import index_by

GROUP_DELTA_LAKE = 'Delta Lake'
GROUP_ORCHESTRATION = 'Orchestration'

TEMPLATES = [
    dict(
        block_type=BlockType.DATA_LOADER,
        description='Load a Delta Table from Amazon S3.',
        groups=[GROUP_DELTA_LAKE],
        language=BlockLanguage.PYTHON,
        name='Amazon S3',
        path='data_loaders/deltalake/s3.py',
    ),
    dict(
        block_type=BlockType.DATA_LOADER,
        description='Load a Delta Table from Azure Blob Storage.',
        groups=[GROUP_DELTA_LAKE],
        language=BlockLanguage.PYTHON,
        name='Azure Blob Storage',
        path='data_loaders/deltalake/azure_blob_storage.py',
    ),
    dict(
        block_type=BlockType.DATA_LOADER,
        description='Load a Delta Table from Google Cloud Storage.',
        groups=[GROUP_DELTA_LAKE],
        language=BlockLanguage.PYTHON,
        name='Google Cloud Storage',
        path='data_loaders/deltalake/gcs.py',
    ),
    dict(
        block_type=BlockType.DATA_LOADER,
        description='Load data from MongoDB.',
        language=BlockLanguage.PYTHON,
        name='MongoDB',
        path='data_loaders/mongodb.py',
    ),
    dict(
        block_type=BlockType.DATA_EXPORTER,
        description='Export data to a Delta Table in Amazon S3.',
        groups=[GROUP_DELTA_LAKE],
        language=BlockLanguage.PYTHON,
        name='Amazon S3',
        path='data_exporters/deltalake/s3.py',
    ),
    dict(
        block_type=BlockType.DATA_EXPORTER,
        description='Export data to a Delta Table in Azure Blob Storage.',
        groups=[GROUP_DELTA_LAKE],
        language=BlockLanguage.PYTHON,
        name='Azure Blob Storage',
        path='data_exporters/deltalake/azure_blob_storage.py',
    ),
    dict(
        block_type=BlockType.DATA_EXPORTER,
        description='Export data to a Delta Table in Google Cloud Storage.',
        groups=[GROUP_DELTA_LAKE],
        language=BlockLanguage.PYTHON,
        name='Google Cloud Storage',
        path='data_exporters/deltalake/gcs.py',
    ),
    dict(
        block_type=BlockType.DATA_EXPORTER,
        description='Export data to MongoDB.',
        language=BlockLanguage.PYTHON,
        name='MongoDB',
        path='data_exporters/mongodb.py',
    ),
    dict(
        block_type=BlockType.DATA_LOADER,
        description='Trigger another pipeline to run.',
        groups=[GROUP_ORCHESTRATION],
        language=BlockLanguage.PYTHON,
        name='Trigger pipeline',
        path='data_loaders/orchestration/triggers/default.jinja',
    ),
    dict(
        block_type=BlockType.DATA_EXPORTER,
        description='Trigger another pipeline to run.',
        groups=[GROUP_ORCHESTRATION],
        language=BlockLanguage.PYTHON,
        name='Trigger pipeline',
        path='data_exporters/orchestration/triggers/default.jinja',
    ),
    dict(
        block_type=BlockType.CALLBACK,
        description='Base template with empty functions.',
        language=BlockLanguage.PYTHON,
        name='Base template',
        path='callbacks/base.jinja',
    ),
    dict(
        block_type=BlockType.CALLBACK,
        description='Trigger another pipeline to run.',
        groups=[GROUP_ORCHESTRATION],
        language=BlockLanguage.PYTHON,
        name='Trigger pipeline',
        path='callbacks/orchestration/triggers/default.jinja',
    ),
    dict(
        block_type=BlockType.CONDITIONAL,
        description='Base template with empty functions.',
        language=BlockLanguage.PYTHON,
        name='Base template',
        path='conditionals/base.jinja',
    ),
]

TEMPLATES_BY_UUID = index_by(lambda x: x['name'], TEMPLATES)
